"""
This script contains the wrapper functions to build a PyKoala RSS object from KOALA (2dfdr-reduced) data.
"""

# =============================================================================
# Basics packages
# =============================================================================
import numpy as np
import os
# =============================================================================
# Astropy and associated packages
# =============================================================================
from astropy import units as u
from astropy.io import fits
from astropy.wcs import WCS
# =============================================================================
# KOALA packages
# =============================================================================
from pykoala import vprint
from pykoala.data_container import RSS

def airmass_from_header(header):
    """
    Compute the airmass extracting the parameters from KOALAS's header'
    """
    # Get ZD, airmass
    ZDSTART = header['ZDSTART']
    ZDEND = header['ZDEND']
    ZD = (ZDSTART + ZDEND) / 2
    airmass = 1 / np.cos(np.radians(ZD))
    return airmass


def koala_header(header):
    """
    Copy 2dfdr headers values from extensions 0 and 2 needed for the initial
    header for PyKoala. (based in the header constructed in  save_rss_fits in
    koala.io)
    """

    # To fit actual PyKoala header format
    header.rename_keyword('CENRA', 'RACEN')
    header.rename_keyword('CENDEC', 'DECCEN')

    cards = [header.cards['BITPIX'],
             header.cards["ORIGIN"],
             header.cards["TELESCOP"],
             header.cards["ALT_OBS"],
             header.cards["LAT_OBS"],
             header.cards["LONG_OBS"],
             header.cards["INSTRUME"],
             header.cards["GRATID"],
             header.cards["SPECTID"],
             header.cards["DICHROIC"],
             header.cards['OBJECT'],
             header.cards["EXPOSED"],
             header.cards["ZDSTART"],
             header.cards["ZDEND"],
             header.cards['NAXIS'],
             header.cards['NAXIS1'],
             header.cards['NAXIS2'],
             header.cards['RACEN'],
             header.cards['DECCEN'],
             header.cards['TEL_PA'],
             header.cards["CTYPE2"],
             header.cards["CUNIT2"],
             header.cards["CTYPE1"],
             header.cards["CUNIT1"],
             header.cards["CRVAL1"],
             header.cards["CDELT1"],
             header.cards["CRPIX1"],
             header.cards["CRVAL2"],
             header.cards["CDELT2"],
             header.cards["CRPIX2"],
             ]
    koala_header = fits.header.Header(cards=cards, copy=False)
    koala_header = header
    return koala_header

def py_koala_fibre_table(fibre_table):
    """
    Generates the spaxels tables needed for PyKoala from the 2dfdr spaxels table.
    """
    # Filtering only selected (in use) fibres
    spaxels_table = fibre_table[fibre_table['SELECTED'] == 1]

    # Defining new arrays
    arr1 = np.arange(len(spaxels_table)) + 1  # +  for starting in 1
    arr2 = np.ones(len(spaxels_table))
    arr3 = np.ones(len(spaxels_table))
    arr4 = np.ones(len(spaxels_table)) * 2048
    arr5 = np.zeros(len(spaxels_table))
    arr6 = spaxels_table['XPOS']
    arr7 = spaxels_table['YPOS']
    arr8 = spaxels_table['SPEC_ID']

    # Defining new columns
    col1 = fits.Column(name='Fibre', format='I', array=arr1)
    col2 = fits.Column(name='Status', format='I', array=arr2)
    col3 = fits.Column(name='Ones', format='I', array=arr3)
    col4 = fits.Column(name='Wavelengths', format='I', array=arr4)
    col5 = fits.Column(name='Zeros', format='I', array=arr5)
    col6 = fits.Column(name='Delta_RA', format='D', array=arr6)
    col7 = fits.Column(name='Delta_Dec', format='D', array=arr7)
    col8 = fits.Column(name='Fibre_OLD', format='I', array=arr8)

    # PyKoala Spaxels table
    py_koala_spaxels_table = fits.BinTableHDU.from_columns([col1, col2, col3, col4, col5, col6, col7, col8])

    return py_koala_spaxels_table

def read_rss(file_path,
             wcs,
             intensity_axis=0,
             variance_axis=1,
             bad_fibres_list=None,
             header=None,
             info=None
             ):
    """TODO."""
    if header is None:
        # Blank Astropy Header object for the RSS header
        # Example how to add header value at the end
        # blank_header.append(('DARKCORR', 'OMIT', 'Dark Image Subtraction'), end=True)
        header = fits.header.Header(cards=[], copy=False)

    file_name = os.path.basename(file_path)

    vprint(f"\n> Reading KOALA RSS file {file_name}")
    #  Open fits file. This assumes that RSS objects are written to file as .fits.
    with fits.open(file_path) as rss_fits:
        # Read intensity using rss_fits_file[0]
        all_intensities = np.array(rss_fits[intensity_axis].data,
                                   dtype=np.float32)
        intensity = np.delete(all_intensities, bad_fibres_list, 0)
        # Bad pixel verbose summary
        vprint(f"No. of fibres in this RSS ={len(all_intensities)}"
               + f"\nNo. of good fibres = {len(intensity)}"
               + f"\nNo. of bad fibres = {len(bad_fibres_list)}")
        if bad_fibres_list is not None:
            vprint(f"Bad fibres = {bad_fibres_list}")
        # Read errors if exist a dedicated axis
        if variance_axis is not None:
            all_variances = rss_fits[variance_axis].data
            variance = np.delete(all_variances, bad_fibres_list, 0)

        else:
            vprint("WARNING! Variance extension not found in fits file!")
            variance = np.full_like(intensity, fill_value=np.nan)

    # Create wavelength from wcs
    nrow, ncol = wcs.array_shape
    wavelength_index = np.arange(ncol)
    #TODO : remove to_value once units are homogeneized
    wavelength = wcs.spectral.array_index_to_world(wavelength_index).to_value("angstrom")
    # First Header value added by the PyKoala routine
    rss = RSS(intensity=intensity << u.adu,
              variance=variance << u.adu**2,
              wavelength=wavelength << u.angstrom,
              info=info,
              header=header,
              fibre_diameter=1.25 << u.arcsec,
              wcs=wcs)

    rss.history('read', ' '.join(['- RSS read from ', file_name]))
    return rss

def koala_rss(path_to_file):
    """
    A wrapper function that converts a file (not an RSS object) to a koala RSS object
    The paramaters used to build the RSS object e.g. bad spaxels, header etc all come from the original (non PyKoala) .fits file
    """
    
    header = fits.getheader(path_to_file, 0) + fits.getheader(path_to_file, 2)
    header = koala_header(header)
    # WCS
    if "RADECSYS" in header:
        header["RADECSYSa"] = header["RADECSYS"]
        del header["RADECSYS"]
    koala_wcs = WCS(header)
    # Fix the WCS information such that koala_wcs.spectra exists
    koala_wcs.wcs.ctype[0] = 'WAVE    '
    # Constructing Pykoala Spaxels table from 2dfdr spaxels table (data[2])
    fibre_table = fits.getdata(path_to_file, 2)
    koala_fibre_table = py_koala_fibre_table(fibre_table)

    # List of bad spaxels from 2dfdr spaxels table
    bad_fibres_list = (fibre_table['SPEC_ID'][fibre_table['SELECTED'] == 0] - 1).tolist()
    # -1 to start in 0 rather than in 1
    # Create the dictionary containing relevant information
    info = {}
    info['name'] = header['OBJECT']
    info['exptime'] = header['EXPOSED'] << u.second
    info['fib_ra'] = (np.rad2deg(header['RACEN'])
                      + koala_fibre_table.data['Delta_RA'] / 3600) << u.deg
    info['fib_dec'] = (np.rad2deg(header['DECCEN'])
                       + koala_fibre_table.data['Delta_DEC'] / 3600) << u.deg
    info['airmass'] = airmass_from_header(header)
    # Read RSS file into a PyKoala RSS object
    rss = read_rss(path_to_file, wcs=koala_wcs,
                   bad_fibres_list=bad_fibres_list,
                   intensity_axis=0,
                   variance_axis=1,
                   header=header,
                   info=info,
                   )
    return rss


def koalaRSS(filename, 
             path=None, 
             rss_object_name = None, 
             plot_map= False, 
             **kwargs):
    """
    Task for creating a KOALA rss object including rss.koala.
    
    This uses koala_rss() wrapper, which reads rss using read_rss()
    
    Then, it adds .koala to the rss object building all the info, see class KOALA_INFO() for rss.
    
    It plots the image (if plot = True) and map (if plot_map = True and plot = True) if requested.
    
    Tasks / Objects that this task uses:
        - KOALA_INFO() : Object keeping the specific KOALA info
        - koala_rss() : Pablo's wrapper read rss for KOALA
        - rss_image() : plotting rss image
        - rss_valid_wave_range() : Computing the rss valid wavelength range (left and right edges for each fibre)
        - compute_integrated_fibre() : computes integrated fibre for a rss
        - get_rss_mask(): Mask for rss
        
    Dictionaries that this task uses:
        - koala_info_rss_dict
        
    Parameters
    ----------
    filename : string
        Name of the rss file.
    path : string, optional
        path to the rss file
    rss_object_name : string, optional
        name of the rss object
    plot_map : boolean, optional
        if True it will plot the rss map. The default is False.
    **kwargs : kwargs
        where we can find plot, verbose, warnings...

    Returns
    -------
    rss : object
        rss KOALA object
    """
    verbose = kwargs.get('verbose', False)
    _warnings_ = kwargs.get('warnings', False)   # We can't use this here, as using warnings.catch_warnings()
    plot =  kwargs.get('plot', False)

    
    if path is not None:
        path_to_file = os.path.join(path,filename)
    else:
        path_to_file = filename  
    
    if verbose: print('\n> Converting KOALA+AAOmega RSS file "'+path_to_file+'" to a koala RSS object...')
    
    # Read rss using standard pykoala  #TODO: Ángel: I think we should merge the 2 of them (koalaRSS and koala_rss)
    rss = koala_rss(path_to_file, verbose=False)   # Using verbose = False to dissable showing the log

    # Create rss.koala
    rss.koala = KOALA_INFO()

    # Now, we add to this rss the information for a koala file
    # Create the dictionary containing relevant information for KOALA
    koala_info = koala_info_rss_dict.copy()  # Avoid missing some key info
    
    koala_info["rss_object_name"] = rss_object_name    # Save the name of the rss object
    koala_info['path_to_file'] = path_to_file          # Save name of the .fits file
    koala_info['n_wave'] = len(rss.wavelength)         # Save n_wave
    koala_info['n_spectra'] = len(rss.intensity)       # Save n_spectra

    # Check that dimensions of fits match KOALA numbers
    if koala_info['n_wave'] != 2048 and koala_info['n_spectra'] != 986: #  1000:
        if _warnings_ or verbose:  print('\n  *** WARNING *** : These numbers are NOT the standard n_wave and n_spectra values for KOALA')
    
    # As now the header is not saved in koala_rss, let's do it again to save here key KOALA info
    header = fits.getheader(path_to_file, 0) + fits.getheader(path_to_file, 2)
    header = koala_header(header)
    
    # Add history
    rss.koala.history = header["HISTORY"]
    
    koala_info['RA_centre_deg'] = header["RACEN"] *180/np.pi
    koala_info['DEC_centre_deg'] = header["DECCEN"] *180/np.pi
    koala_info['exptime'] = header["EXPOSED"]
    
    # Get AAOmega Arm & gratings
    if (header['SPECTID'] == "RD"):      
        koala_info['aaomega_arm'] = "red"
    if (header['SPECTID'] == "BL"):      
        koala_info['aaomega_arm'] = "blue"    
    koala_info['aaomega_grating'] = header['GRATID']
    koala_info['aaomega_dichroic'] = header["DICHROIC"]
    
    # Constructing Pykoala Spaxels table from 2dfdr spaxels table (data[2]) AGAIN to know WIDE/NARROW & spaxel_size
    fibre_table = fits.getdata(path_to_file, 2)
    koala_fibre_table = py_koala_fibre_table(fibre_table)
    
    fib_ra_offset = koala_fibre_table.data['Delta_RA'] 
    fib_dec_offset = koala_fibre_table.data['Delta_DEC'] 
    # Check if NARROW (spaxel_size = 0.7 arcsec) or WIDE (spaxel_size=1.25) field of view
    # (if offset_max - offset_min > 31 arcsec in both directions)
    if (np.max(fib_ra_offset) - np.min(fib_ra_offset)) > 31 or (np.max(fib_dec_offset) - np.min(fib_dec_offset)) > 31:
        spaxel_size = 1.25
        KOALA_fov = 'WIDE: 50.6" x 27.4"'
    else:
        spaxel_size = 0.7
        KOALA_fov = 'NARROW: 28.3" x 15.3"'   
    koala_info['KOALA_fov']=KOALA_fov
    koala_info['spaxel_size']=spaxel_size 
    koala_info['position_angle'] = header['TEL_PA']
    
    # Check valid range (basic mask in PyKOALA)
    valid_wave_range_data = rss_valid_wave_range (rss)    
    koala_info['valid_wave_min'] = valid_wave_range_data[2][0]
    koala_info['valid_wave_max'] = valid_wave_range_data[2][1]
    koala_info['valid_wave_min_index'] = valid_wave_range_data[1][0]
    koala_info['valid_wave_max_index'] = valid_wave_range_data[1][1]   
    # koala_info['first_good_wave_per_fibre'] = np.array(valid_wave_range_data[0][0])
    # koala_info['last_good_wave_per_fibre'] = np.array(valid_wave_range_data[0][1])
    # koala_info['list_fibres_all_good_values'] = np.array(valid_wave_range_data[3])
    rss.koala.mask = [np.array(valid_wave_range_data[0][0]), np.array(valid_wave_range_data[0][1]) ]
    rss.koala.list_fibres_all_good_values = np.array(valid_wave_range_data[3])
        
    # Saving the info to the rss object
    rss.koala.info = koala_info
    rss.koala.header = header                     # Saving header
    rss.koala.fibre_table = koala_fibre_table           # Saving original koala fibre table as needed later

    if "RADECSYS" in header:
        header["RADECSYSa"] = header["RADECSYS"]
        del header["RADECSYS"]
    rss.koala.wcs = WCS(header)   # Saving WCS, not sure if later needed, but for now it is there
    
    # Plot RSS image if requested  #!!!
    if plot: rss_image(rss, **kwargs)
    
    # Computing integrating time and plot map if requested (plot_map in **kwargs is used in compute_integrated_fibre):
    kwargs["plot"] = plot_map
    verbose = kwargs.get('verbose', False)
    kwargs["verbose"] = False
    rss=compute_integrated_fibre(rss, **kwargs)
    kwargs["verbose"] = verbose
     
    #get mask
    rss.mask = get_rss_mask(rss, **kwargs)
    
    # Printing the info is requested:
    if verbose:    
        print('  Found {} spectra with {} wavelengths'.format(koala_info['n_spectra'], koala_info['n_wave']),
                      'between {:.2f} and {:.2f} Angstroms.'.format(rss.wavelength[0], rss.wavelength[-1]))
        print('  This RSS file uses the',koala_info['aaomega_grating'],'grating in the',koala_info['aaomega_arm'],'AAOmega arm.')
        print('  The KOALA field of view is {}, with a spaxel size of {}" and PA = {:.1f}º.'.format(KOALA_fov, spaxel_size,koala_info['position_angle']))
    
        if koala_info['rss_object_name'] is not None: 
            print('  Name of the observation = "{}",   Name of this Python RSS object = "{}".'.format(rss.info['name'],koala_info['rss_object_name']))
        else:
            print('  Name of the observation = "{}".'.format(rss.info['name']))    
        
    if kwargs.get("description") is not None:
        rss.koala_info["description"] = kwargs.get("description")
        if verbose: print('  Description provided to this KOALA RSS object = "{}".'.format(koala_info['description']))
            
    return rss
# #-----------------------------------------------------------------------------
# %% ==========================================================================
# =============================================================================
def compute_integrated_fibre(rss, 
                             list_spectra=None, 
                             valid_wave_min=None,
                             valid_wave_max=None, 
                             min_value_to_plot=0.01,
                             title=" - Integrated values",
                             text="...",
                             correct_negative_sky=False,
                             order_fit_negative_sky=3,
                             kernel_negative_sky=51, low_fibres=10,
                             individual_check=True,
                             use_fit_for_negative_sky=False,
                             last_check=False,
                             **kwargs): #plot=False, warnings=True, verbose=True

    """
    Compute the integrated flux of a fibre in a particular range

    Parameters
    ----------
    list_spectra: float (default "all")
        list with the number of fibres for computing integrated value
        if using "all" it does all fibres
    valid_wave_min, valid_wave_max :  float
        the integrated flux value will be computed in the range [valid_wave_min,valid_wave_max]
        (default = , if they all 0 we use [self.valid_wave_min,self.valid_wave_max]
    min_value_to_plot: float (default 0)
        For values lower than min_value, we set them as min_value
    title : string
        Tittle for the plot
    text: string
        A bit of extra text
    correct_negative_sky : boolean (default = False)
        Corrects negative values making 0 the integrated flux of the lowest fibre
    last_check: boolean (default = False)
        If that is the last correction to perform, say if there is not any fibre
        with has an integrated value < 0.
    **kwargs : kwargs
        where we can find verbose, warnings, plot...

    Example
    ----------
    star1 = compute_integrated_fibre(star1,valid_wave_min=6500, valid_wave_max=6600,
    title = " - [6500,6600]", plot = True)
    """
    verbose = kwargs.get('verbose', False)
    plot =  kwargs.get('plot', False)
    warnings = kwargs.get('warnings', verbose)
    
    wavelength = rss.wavelength
    n_spectra = len(rss.intensity)
    
    if list_spectra is None: list_spectra = list(range(n_spectra))
    if valid_wave_min is None:  valid_wave_min = rss.koala.info["valid_wave_min"]        
    if valid_wave_max is None:  valid_wave_max = rss.koala.info["valid_wave_max"]

        
    if verbose: print("\n> Computing integrated fibre values in range [ {:.2f} , {:.2f} ] {}".format(valid_wave_min, valid_wave_max, text))


    region = np.where((wavelength > valid_wave_min
                       ) & (wavelength < valid_wave_max))[0]
    waves_in_region = len(region)

    integrated_fibre = np.nansum(rss.intensity[:, region], axis=1)
    integrated_fibre_variance = np.nansum(rss.variance[:, region], axis=1)
    negative_fibres = (np.where(integrated_fibre < 0)[0]).tolist()
    n_negative_fibres = len(negative_fibres)    # n_negative_fibres = len(integrated_fibre[integrated_fibre < 0])

    if verbose:
        print("  - Median value of the integrated flux =", np.round(np.nanmedian(integrated_fibre), 2))
        print("                                    min =", np.round(np.nanmin(integrated_fibre), 2), ", max =",
              np.round(np.nanmax(integrated_fibre), 2))
        print("  - Median value per wavelength         =",
              np.round(np.nanmedian(integrated_fibre) / waves_in_region, 2))
        print("                                    min = {:9.3f} , max = {:9.3f}".format(
            np.nanmin(integrated_fibre) / waves_in_region, np.nanmax(integrated_fibre) / waves_in_region))

    if len(negative_fibres) != 0:
        if warnings or verbose: print(
            "\n> WARNING! : Number of fibres with integrated flux < 0 : {}, that is the {:5.2f} % of the total !".format(
                n_negative_fibres, n_negative_fibres * 100. / n_spectra))
        if correct_negative_sky:
            if verbose: print("  CORRECTING NEGATIVE SKY HERE NEEDS TO BE IMPLEMENTED - NOTHING ELSE DONE")
            pass  ## NEDD TO IMPLEMENT THIS #TODO
            # rss.correcting_negative_sky(plot=plot, order_fit_negative_sky=order_fit_negative_sky,
            #                              individual_check=individual_check,
            #                              kernel_negative_sky=kernel_negative_sky,
            #                              use_fit_for_negative_sky=use_fit_for_negative_sky, low_fibres=low_fibres)
        else:
            if plot and verbose:
                print(
                    "\n> Adopting integrated flux = {:5.2f} for all fibres with negative integrated flux (for presentation purposes)".format(
                        min_value_to_plot))
                print("  This value is {:5.2f} % of the median integrated flux per wavelength".format(
                    min_value_to_plot * 100. / np.nanmedian(integrated_fibre) * waves_in_region))
                integrated_fibre_plot = integrated_fibre.clip(min=min_value_to_plot, max=integrated_fibre.max())
    else:
        integrated_fibre_plot = integrated_fibre
        if last_check:
            if warnings or verbose: print("\n> There are no fibres with integrated flux < 0 !")

    integrated_fibre_sorted = np.argsort(integrated_fibre).tolist()
    if plot: rss_map(rss,  variable=integrated_fibre_plot, title=title , **kwargs)
    
    rss.koala.integrated_fibre = integrated_fibre
    rss.koala.integrated_fibre_variance = integrated_fibre_variance
    rss.koala.integrated_fibre_sorted = integrated_fibre_sorted
    rss.koala.negative_fibres = negative_fibres
    
    return rss



# Mr Krtxo \(ﾟ▽ﾟ)/ + Ángel :-)



"""Module to create mock datacontainers"""

# =============================================================================
# Basics packages
# =============================================================================
import numpy as np
import os
# =============================================================================
# Astropy and associated packages
# =============================================================================
from astropy import units as u
from astropy.io import fits
from astropy.wcs import WCS
# =============================================================================
# pyKOALA packages
# =============================================================================
from pykoala import vprint
from pykoala.data_container import RSS
from pykoala.ancillary import check_unit


class NoiseModel:
    """Noise model for mock data."""
    def __init__(self, bias_bckgr, gaussian_sigma, poisson_flux_thresh, poisson_sigma):
        self.bias_bckgr = np.asarray(bias_bckgr)
        self.gaussian_sigma = gaussian_sigma
        self.poisson_flux_thresh = poisson_flux_thresh
        self.poisson_sigma = poisson_sigma

    def gaussian_noise(self, intensity):
        return np.random.normal(0, self.gaussian_sigma, size=intensity.shape)

    def poisson_noise(self, intensity):
        return self.poisson_sigma * np.sqrt(intensity / self.poisson_flux_thresh)

    def __call__(self, intensity):
        if (self.bias_bckgr.ndim > 0) and (self.bias_bckgr.shape != intensity.shape):
            raise ArithmeticError(f"Bias background dimensions ({self.bias_bckgr}) do not"
                                  f"match intensity shape {intensity.shape}")
        return (self.bias_bckgr + self.gaussian_noise(intensity)
                + self.poisson_noise(intensity))

def gaussian_source(fibre_ra, fibre_dec, source_ra, source_dec,
                    source_ra_sigma, source_dec_sigma, source_intensity=1):
    x = (fibre_ra - source_ra) / source_ra_sigma
    y = (fibre_dec - source_dec) / source_dec_sigma
    if x.ndim != source_intensity.ndim:
        x = x[:, np.newaxis]
        y = y[:, np.newaxis]
    intensity =  np.exp(- 0.5 * x**2 - 0.5 * y**2) * source_intensity
    return intensity

def mock_rss(ra_n_fibres=20, dec_n_fibres=20, fibre_diameter=1.5 << u.arcsec,
             fibre_separation=0 << u.arcsec,
             n_wave=1000, wave_range=(3000 << u.AA, 8000 << u.AA),
             ra_cen=180 << u.deg, dec_cen=45 << u.deg, exptime=600 << u.second,
             airmass=1.0,
             source_kwargs={},
             noise_kwargs={},
             nan_frac=0.0):
    """Create a mock RSS object."""
    vprint("Creating mock RSS data")
    n_fibres = ra_n_fibres * dec_n_fibres
    fibre_diameter = check_unit(fibre_diameter)
    fibre_separation = check_unit(fibre_separation)
    fov = (ra_n_fibres * (fibre_diameter + fibre_separation) - fibre_separation,
           dec_n_fibres * (fibre_diameter + fibre_separation) - fibre_separation)

    ra = np.linspace(- fov[0] / 2, fov[0] / 2, ra_n_fibres) + ra_cen
    dec = np.linspace(- fov[1] / 2, fov[1] / 2, dec_n_fibres) + dec_cen
    ra, dec = np.meshgrid(ra, dec)
    vprint(f"Total number of fibres: {n_fibres}")

    # Ensure units consistency
    wave_range = [check_unit(wr, u.AA) for wr in wave_range]
    wavelength = np.linspace(*wave_range, n_wave)

    # Create intensity and variance attributes
    intensity = np.zeros((n_fibres, wavelength.size))
    variance = np.zeros_like(intensity)
   
    # Add sources and noise
    if "source_ra" not in source_kwargs:
        source_kwargs["source_ra"] = ra_cen
    if "source_dec" not in source_kwargs:
        source_kwargs["source_dec"] = dec_cen
    if "source_ra_sigma" not in source_kwargs:
        source_kwargs["source_ra_sigma"] = fov[0] / 2
    if "source_dec_sigma" not in source_kwargs:
        source_kwargs["source_dec_sigma"] = fov[1] / 2
    if "source_intensity" not in source_kwargs:
        source_kwargs["source_intensity"] = 10 * np.ones_like(intensity)

    intensity += gaussian_source(ra.flatten(), dec.flatten(), **source_kwargs)
    
    # Add noise
    if "bias_bckgr" not in noise_kwargs:
        noise_kwargs["bias_bckgr"] = 0.05
    if "gaussian_sigma" not in noise_kwargs:
        noise_kwargs["gaussian_sigma"] = 0.1
    if "poisson_flux_thresh" not in noise_kwargs:
        noise_kwargs["poisson_flux_thresh"] = 0.75
    if "poisson_sigma" not in noise_kwargs:
        noise_kwargs["poisson_sigma"] = 0.1

    noise_model = NoiseModel(**noise_kwargs)
    intensity += noise_model(intensity)
    variance += noise_kwargs["gaussian_sigma"]**2

    # Add nans (e.g. masked cosmic rays)
    if nan_frac > 0:
        nan_idx = np.random.randint(0, intensity.size, size=int(intensity.size * nan_frac))
        intensity[np.unravel_index(nan_idx, intensity.shape)] = np.nan
        variance[np.unravel_index(nan_idx, intensity.shape)] = np.nan

    # Create the RSS object
    info = {}
    info['name'] = "MockRSS"
    info['exptime'] = exptime
    info['fib_ra'] = ra.flatten()
    info['fib_dec'] = dec.flatten()
    info['airmass'] = airmass

    rss = RSS(intensity=intensity << u.adu,
              variance=variance << u.adu**2,
              wavelength=wavelength,
              info=info,
              fibre_diameter=fibre_diameter,
              wcs=WCS())

    rss.history('Mock', "Mock RSS generated by pykoala")
    return rss

if __name__ == "__main__":
    rss = mock_rss()
    
    rss.plot_fibres()




# =============================================================================
# Basics packages
# =============================================================================
import numpy as np

from pykoala import vprint
from pykoala.data_container import RSS

# =============================================================================
# Combine RSS (e.g., flats, twilights)
# =============================================================================

def combine_rss(list_of_rss, combine_method='nansum'):
    """Combine an input list of DataContainers into a new DataContainer.
    
    This method combines the intensity and variance of an input list of RSS
    frames.

    Parameters
    ----------
    list_of_rss : list
        A list containing the input RSS
    
    Returns
    -------
    stacked_rss : :class:`RSS`
        The resulting RSS from combining all input intensities and variances.
    """
    vprint("Combining input list of RSS data")
    all_intensities = []
    all_variances = []
    finite_masks = []

    for rss in list_of_rss:
        intensity = rss.intensity.copy()
        variance = rss.variance.copy()

        # Ensure nans
        finite_mask = np.isfinite(intensity) & np.isfinite(variance)
        finite_masks.append(finite_mask)

        intensity[~finite_mask] = np.nan
        variance[~finite_mask] = np.nan

        all_intensities.append(intensity)
        all_variances.append(variance)

    if hasattr(np, combine_method):
        combine_function = getattr(np, combine_method)
        new_intensity = combine_function(all_intensities, axis=0
                                         ) << rss.intensity.unit
        new_variance = combine_function(all_variances, axis=0
                                        ) << rss.variance.unit
    else:
        raise NotImplementedError("Implement user-defined combining methods")
    # TODO: Update the metadata as well
    new_rss = RSS(intensity=new_intensity, variance=new_variance,
                  wavelength=rss.wavelength, history=rss.history, info=rss.info,
                  fibre_diameter=rss.fibre_diameter, sky_fibres=rss.sky_fibres)
    return new_rss


# =============================================================================
# Mr Krtxo \(ﾟ▽ﾟ)/
#                                                       ... Paranoy@ Rulz! ;^D



"""
Utility functions.

Many of them are partially implemented. They were in the
old, non-modular version of PyKOALA and have not been included
in the current modular scheme.
"""

# =============================================================================
# Basics packages
# =============================================================================
import os
import numpy as np
from matplotlib import pyplot as plt
from scipy import interpolate
from scipy import optimize
from shapely import geometry
# =============================================================================
# Astropy and associated packages
# =============================================================================
from astropy.io import fits
from astropy import units as u
from astropy.modeling.models import custom_model
# =============================================================================

# =============================================================================
# PyKOALA modules
# =============================================================================
from pykoala import vprint


def check_unit(quantity, default_unit=None, equivalencies=[]):
    """Check the units of an input quantity.
    
    Parameters
    ----------
    quantity : np.ndarray or astropy.units.Quantity
        Input quantity.
    default_unit : astropy.units.Quantity, default=None
        If `quantity` has not units, it corresponds to the unit assigned to it.
        Otherwise, it is used to check the equivalency with `quantity`.
    
    Returns
    -------
    quantity: :class:`astropy.units.Quantity`
        Converted quantity.
    """
    if quantity is None:
        return quantity
    isq = isinstance(quantity, u.Quantity)
    if isq and default_unit is not None:
        if not quantity.unit.is_equivalent(default_unit,
                                           equivalencies=equivalencies):
            raise u.UnitTypeError(
                "Input quantity does not have the appropriate units")
        else:
            return quantity.to(default_unit, equivalencies=equivalencies)
    elif not isq and default_unit is not None:
        return quantity * default_unit
    elif not isq and default_unit is None:
        raise ValueError("Input value must be a astropy.units.Quantity")
    else:
        return quantity

def remove_unit(quantity, default_unit=None):
    """Convert an :class:`astropy.units.Quantity` into a :class:`numpy.array`.
    
    This method converts an input quantity into a an array eighter by taking
    the value associated to the current units, or after converting the quantity
    into the input units.

    Parameters
    ----------
    quantity : np.ndarray or astropy.units.Quantity
        Input quantity.
    default_unit : astropy.units.Quantity, default=None
        If `quantity` has not units, it corresponds to the unit assigned to it.
        Otherwise, it is used to check the equivalency with `quantity`.
    
    Returns
    -------
    array: :class:`numpy.array`
        Array associated to the input quantity.
    """
    isq = isinstance(quantity, u.Quantity)
    if isq and default_unit is not None:
        if not quantity.unit.is_equivalent(default_unit):
            raise u.UnitTypeError(
                "Input quantity does not have the appropriate units")
        else:
            return quantity.to_value(default_unit)
    elif not isq:
        return quantity
    else:
        return quantity.value

def preserve_units_dec(func):
    """Decorator method to preserve `astropy.Units` on input arguments."""
    def wrapper(data, *args, **kwargs):
        if isinstance(data, u.Quantity):
            unit = data.unit
        else:
            unit = 1
        val = func(data, *args, **kwargs)
        
        if not isinstance(val, u.Quantity):
            return val << unit
        else:
            return val
    return wrapper

def remove_units_dec(func):
    """Decorator method to remove `astropy.Units` from input arguments."""
    def wrapper(*args, **kwargs):
        unitless_args = [remove_unit(a) for a in args]
        unitless_kwargs = {k : remove_unit(v) for k, v in kwargs.items()}
        return func(*unitless_args, **unitless_kwargs)
    return wrapper


def detect_edge(rss):
    """
    Detect the edges of a RSS. Returns the minimum and maximum wavelength that 
    determine the maximum interval with valid (i.e. no masked) data in all the 
    spaxels.

    Parameters
    ----------
    rss : RSS object.

    Returns
    -------
    min_w : float
        The lowest value (in units of the RSS wavelength) with 
        valid data in all spaxels.
    min_index : int
        Index of min_w in the RSS wavelength variable.
    max_w : float
        The higher value (in units of the RSS wavelength) with 
        valid data in all spaxels.
    max_index : int
        Index of max_w in the RSS wavelength variable.

    """
    collapsed = np.sum(rss.intensity, 0)
    nans = np.isfinite(collapsed)
    wavelength = rss.wavelength
    min_w = wavelength[nans].min()
    min_index = wavelength.tolist().index(min_w)
    max_w = wavelength[nans].max()
    max_index = wavelength.tolist().index(max_w)
    return min_w, min_index, max_w, max_index

# ----------------------------------------------------------------------------------------------------------------------
# WCS operations
# ----------------------------------------------------------------------------------------------------------------------


def update_wcs_coords(wcs, ra_dec_val=None, ra_dec_offset=None):
    """Update the celestial reference values of a WCS.

    Update the celestial coordinates of a WCS using new central values of RA
    and DEC or relative offsets expressed in degree.

    Parameters
    ----------
    - wcs: asteropy.wcs.WCS
        Target WCS to update.
    - ra_dec_val: list or tupla, default=None
        New CRVAL of RA and DEC. Both elements must be instances of
        :class:`astropy.units.Quantity`.
    - ra_dec_offset: list or tupla, default=None
        Relative offset that will be applyied to CRVAL of RA and DEC axis. If
        `ra_dec_val` is privided, this will be ignored. Both elements must be
        instances of :class:`astropy.units.Quantity`.

    Return
    ------
    - correct_wcs: astropy.wcs.WCS
        A copy of the original WCS with the reference values updated.
    """
    correc_wcs = wcs.deepcopy()
    if ra_dec_val is not None:
        if "RA" in correc_wcs.wcs.ctype[0]:
            correc_wcs.wcs.crval[0] = ra_dec_val[0].to_value(
                correc_wcs.wcs.cunit[0])
            correc_wcs.wcs.crval[1] = ra_dec_val[1].to_value(
                correc_wcs.wcs.cunit[1])
        elif "RA" in correc_wcs.wcs.ctype[1]:
            correc_wcs.wcs.crval[0] = ra_dec_val[1].to_value(
                correc_wcs.wcs.cunit[0])
            correc_wcs.wcs.crval[1] = ra_dec_val[0].to_value(
                correc_wcs.wcs.cunit[1])
        else:
            raise NameError(
                "RA coordinate could not be found in the WCS coordinate types:"
                + f"{correc_wcs.wcs.ctype[0]}, {correc_wcs.wcs.ctype[1]}")
    elif ra_dec_offset is not None:
        if "RA" in correc_wcs.wcs.ctype[0]:
            correc_wcs.wcs.crval[0] = correc_wcs.wcs.crval[0] + \
                ra_dec_offset[0].to_value(correc_wcs.wcs.cunit[0])
            correc_wcs.wcs.crval[1] = correc_wcs.wcs.crval[1] + \
                ra_dec_offset[1].to_value(correc_wcs.wcs.cunit[1])
        elif "RA" in correc_wcs.wcs.ctype[1]:
            correc_wcs.wcs.crval[0] = correc_wcs.wcs.crval[0] + \
                ra_dec_offset[1].to_value(correc_wcs.wcs.cunit[0])
            correc_wcs.wcs.crval[1] = correc_wcs.wcs.crval[1] + \
                ra_dec_offset[0].to_value(correc_wcs.wcs.cunit[1])
        else:
            raise NameError(
                "RA coordinate could not be found in the WCS coordinate types:"
                + f"{correc_wcs.wcs.ctype[0]}, {correc_wcs.wcs.ctype[1]}")

    return correc_wcs


# ----------------------------------------------------------------------------------------------------------------------
# Arithmetic operations
# ----------------------------------------------------------------------------------------------------------------------

def med_abs_dev(x, axis=0):
    """Compute the Median Absolute Deviation (MAD) from an input array.
    
    Parameters
    ----------
    x : :class:`np.ndarray`
        Input data.
    axis : int of tupla, optional
        Array axis along with the MAD will be computed
    
    Returns
    -------
    mad : np.ndarray
        Associated MAD to x along the chosen axes.
    """
    mad = np.nanmedian(
        np.abs(x - np.expand_dims(np.nanmedian(x, axis=axis), axis=axis)),
        axis=axis)
    return mad


def std_from_mad(x, axis=0):
    """Estimate the estandard deviation from the MAD.

    Parameters
    ----------
    x : :class:`np.ndarray`
        Input data.
    axis : int of tupla, optional
        Array axis along with the MAD will be computed

    Returns
    -------
    mad : np.ndarray
        Associated MAD to x along the chosen axes.
    
    See also
    --------
    :func:`med_abs_dev`
    """
    return 1.4826 * med_abs_dev(x, axis=axis)

def running_mean(x, n_window):
    """
    This function calculates the running mean of an array.

    Parameters
    ----------
    x : (n,) np.ndarray
        This is the given array.
    n_window : Int
        Number of neighbours for computing the running mean.

    Returns
    -------
    running_mean_x: (n - n_window,) np.ndarray
        Running mean array.
    """
    cumsum = np.cumsum(np.insert(x, 0, 0))
    return (cumsum[n_window:] - cumsum[:-n_window]) / n_window

def symmetric_background(x, weights=None, fig_name=None):
    """
    This function calculates the mode of an array,
    assuming that the data can be decomposed as:
    signal plus uniform background plus noise with a symmetric
    probability distribution.

    Parameters
    ----------
    x : np.ndarray
        Data.
    weigths : np.ndarray (None)
        Statistical weight of each point (defaults to None; equal weights).
    fig_name : str (None)
        Name of the figure to be plotted, if desired.

    Returns
    -------
    mode : x.dtype
        Mode of the input array.
    threshold : x.dtype
        Optimal separation between signal and noise background.
    """
    if weights is None:
        weights = np.ones_like(x).astype(float)

    valid = np.where(np.isfinite(x) & np.isfinite(weights))
    n_valid = valid[0].size
    if not n_valid > 2:
        return np.nan, np.nan

    sorted_by_x = np.argsort(x[valid].flatten())
    sorted_x = x[valid].flatten()[sorted_by_x]
    sorted_weight = weights[valid].flatten()[sorted_by_x]

    cumulative_mass = np.cumsum(sorted_weight)
    total_mass = cumulative_mass[-1]
    sorted_weight /= total_mass
    cumulative_mass /= total_mass
    
    
    nbins = int(2 + np.sqrt(n_valid))
    m_left = np.linspace(0, 0.5, nbins)[1:-1]
    m_mid = 2 * m_left
    m_right = 0.5 + m_left

    x_left = np.interp(m_left, cumulative_mass, sorted_x)
    x_mid = np.interp(m_mid, cumulative_mass, sorted_x)
    x_right = np.interp(m_right, cumulative_mass, sorted_x)

    h = np.fmin(x_right-x_mid, x_mid-x_left)
    valid = np.where(h > 0)
    if not valid[0].size > 0:
        return np.nan, np.nan
    x_mid = x_mid[valid]
    h = h[valid]
    rho = (np.interp(x_mid+h, sorted_x, cumulative_mass) - np.interp(x_mid-h, sorted_x, cumulative_mass)) /2/h

    rho_threshold = np.nanpercentile(rho, 100*(1 - 1/np.sqrt(nbins)))
    peak_region = x_mid[rho >= rho_threshold]
    index_min = np.searchsorted(sorted_x, np.min(peak_region))
    index_max = np.searchsorted(sorted_x, np.max(peak_region))
    index_mode = (index_min+index_max) // 2
    mode = sorted_x[index_mode]
    m_mode = cumulative_mass[index_mode]

    rho_bg = np.fmin(rho, np.interp(x_mid, (2*mode - x_mid)[::-1], rho[::-1], left=0, right=0))
    if m_mode <= 0.5:
        total_bg = 2 * m_mode
        threshold = sorted_x[np.clip(int(n_valid * total_bg), 0, n_valid-1)]
        contamination = np.interp(2*mode-threshold, sorted_x, cumulative_mass)
    else:
        total_bg = 2 * (1 - m_mode)
        threshold = sorted_x[np.clip(int(n_valid * (1-total_bg)), 0, n_valid-1)]
        contamination = 1 - np.interp(2*mode-threshold, sorted_x, cumulative_mass)

    if fig_name is not None:
        plt.close(fig_name)
        fig = plt.figure(fig_name, figsize=(8, 5))
        axes = fig.subplots(nrows=1, ncols=1, squeeze=False,
                            sharex='col', sharey='row',
                            gridspec_kw={'hspace': 0, 'wspace': 0},
                           )

        ax = axes[0, 0]
        ax.set_ylabel('probability density')
        #ax.set_yscale('log')
        ax.plot(x_mid, rho-rho_bg, 'b-', alpha=.5, label='p(signal)')
        ax.plot(x_mid, rho_bg, 'r-', alpha=.5, label='p(background)')
        ax.plot(x_mid, rho, 'k-', alpha=.5, label='total')


        ax.set_xlabel('value')
        L = 5 * np.abs(threshold - mode)
        vmin = np.max([mode - L, x_mid[0]])
        vmax = np.min([mode + L, x_mid[-1]])
        ax.set_xlim(vmin, vmax)

        for ax in axes.flatten():
            ax.tick_params(which='both', bottom=True, top=True, left=True, right=True)
            ax.tick_params(which='major', direction='inout', length=8, grid_alpha=.3)
            ax.tick_params(which='minor', direction='in', length=2, grid_alpha=.1)
            ax.grid(True, which='both')
            ax.axvspan(sorted_x[index_min], sorted_x[index_max], color='k', alpha=.1)
            ax.axvline(mode, c='k', ls=':', label=f'background = {mode:.4g}')
            ax.axvline(threshold, c='b', ls='-.', label=f'signal threshold = {threshold:.4g}')
            ax.axvline(2*mode - threshold, c='r', ls='-.', alpha=.5, label=f'contamination = {100*contamination/(1-total_bg):.2f}%')
            ax.legend()

        fig.suptitle(fig_name)
        fig.set_tight_layout(True)
        return mode, threshold, fig
        
    return mode, threshold

def parabolic_maximum(x, f):
    """Find the maximum by fitting a parabolic function to three points.
    
    Parameters
    ----------
    x : np.array
        Values where the parabola is evaluated
    y : np.array
        Input function values
    
    Returns
    -------
    x_max
    """
    if x.size != 3:
        raise ValueError("Sice of x and f must be 3")
    f23 = f[1] - f[2]
    f21 = f[1] - f[0]
    x21 = x[1] - x[0]
    x23 = x[1] - x[2]
    x_max = x[1] - 0.5 * (x21**2 * f23 - x23**2 * f21) / (x21 * f23 - x23 * f21)
    return x_max

def flux_conserving_interpolation(new_wave : u.Quantity, wave : u.Quantity,
                                  spectra : u.Quantity) -> u.Quantity:
    """Interpolate a spectra to a new grid of wavelengths preserving the flux density.
    
    Parameters
    ----------
    new_wave : :class:`np.ndarray` or :class:`astropy.units.Quantiy`
        New grid of wavelengths
    wave : np.ndarray
        Original grid of wavelengths
    spectra : :class:`astropy.units.Quantity`
        Spectra associated to `wave`.
    
    Returns
    -------
    interp_spectra : np.ndarray
        Interpolated spectra to `new_wave`
    """
    wave = check_unit(wave)
    new_wave = check_unit(new_wave, wave.unit)
    # Strict check
    # Spectra can have different, non-compatible units, such as ADU or flam
    spectra = check_unit(spectra)
    mask = np.isfinite(spectra)
    masked_wave = wave[mask]

    wave_limits = 1.5 * masked_wave[[0, -1]] - 0.5 * masked_wave[[1, -2]]
    wave_edges = np.hstack(
        [wave_limits[0],
         (masked_wave[1:] + masked_wave[:-1])/2,
         wave_limits[1]])

    new_wave_limits = 1.5 * new_wave[[0, -1]] - 0.5 * new_wave[[1, -2]]
    new_wave_edges = np.hstack(
        [new_wave_limits[0],
         (new_wave[1:] + new_wave[:-1])/2,
         new_wave_limits[1]])
    cumulative_spectra = np.cumsum(spectra[mask] * np.diff(wave_edges))
    cumulative_spectra = np.insert(cumulative_spectra, 0,
                                   0 << cumulative_spectra.unit)
    new_cumulative_spectra = np.interp(new_wave_edges, wave_edges,
                                       cumulative_spectra)
    interp_spectra = np.diff(new_cumulative_spectra) / np.diff(new_wave_edges)
    return interp_spectra


def centre_of_mass(w, x, y):
    """Compute the centre of mass from a distribution of points and weights.

    Parameters
    ----------
    x: np.ndarray
        Coordinates corresponding to the x-axis.
    y: np.ndarray
        Coordinates corresponding to the y-axis.
    w: np.ndarray
        Weights for computing the centre of mass.

    Returns
    -------
    center_of_mass : tupla
        Center of mass expressed as ``(x_com, y_com)``
    """
    norm = np.nansum(w)
    x_com, y_com = np.nansum(w * x) / norm, np.nansum(w * y) / norm
    if np.isfinite(x_com) and np.isfinite(y_com):
        return x_com, y_com
    else:
        raise RuntimeError(
            "Failed computing centre of mass computed for\n w={}\n x={}\n y={}"
            .format(w, x, y))

# TODO: Stale method
def growth_curve_1d(f, x, y):
    """TODO"""
    r2 = x**2 + y**2
    idx_sorted = np.argsort(r2)
    growth_c = np.nancumsum(f[idx_sorted])
    return r2[idx_sorted], growth_c

# TODO: Stale method
def growth_curve_2d(image, x0=None, y0=None):
    """Compute the curve of growth of an array f with respect to a given point (x0, y0).

    Parameters
    ----------
    image: np.ndarray
        2D array
    x0: float
        Origin of coordinates in pixels along the x-axis (columns).
    y0: float
        Origin of coordinates in pixels along the y-axis (rows).

    Returns
    -------
    r2: np.ndarray
        Vector containing the square radius with respect to (x0, y0).
    growth_curve: np.ndarray
        Curve of growth centered at (x0, y0).
    """
    xx, yy = np.meshgrid(np.arange(0, image.shape[1]),
                         np.arange(0, image.shape[1]))
    r2 = (xx - x0) ** 2 + (yy - y0) ** 2
    idx_sorted = np.argsort(r2)
    growth_c = np.cumsum(image.flatten()[idx_sorted])
    return r2[idx_sorted], growth_c

@preserve_units_dec
def interpolate_image_nonfinite(image):
    """Use :class:`scipy.interpolate.NearestNDInterpolator` to replace NaN values.

    Parameters
    ----------
    - image: :class:`np.ndarray`
        2D array to be interpolated
    Returnrs
    --------
    - interpolated_image: :class:`np.ndarray`
        Image with nan values replaced by their nearest-neightbour values.
    """
    if image.ndim != 2:
        raise ArithmeticError(f"Input image must have 2D not {image.ndim}")

    x, y = np.meshgrid(
        np.arange(0, image.shape[1], 1),
        np.arange(0, image.shape[0], 1))
    mask = np.isfinite(image)
    if not mask.any():
        raise ArithmeticError("All values of input image are non-finite")
    interp = interpolate.NearestNDInterpolator(
        list(zip(x[mask], y[mask])), image[mask].value)
    interp_image = interp(x, y)
    return interp_image

def vac_to_air(vac_wl: u.Quantity):
    """Convert wavelength in vacuum to air using Morton (1991, ApJS, 77, 119).
    
    Parameters
    ----------
    - vac_wl: :class:`astropy.units.Quantity`
        Vector of vacuum wavelengths in Angstrom.
    
    Returns
    -------
    - air_wl: :class:`astropy.units.Quantity`
        Vector of air wavelengths in Angstrom
    """
    sigma = 1 / vac_wl.to_value("micron")
    vac_over_air = (1 + 8.0605e-5 + 2.48099e-2 / (132.274 - sigma**2)
                    + 1.74557e-4 / (39.32957 - sigma**2)
                    ) << u.dimensionless_unscaled
    return vac_wl / vac_over_air

# TODO: stale
def smooth_spectrum(wlm, s, wave_min=0, wave_max=0, step=50, exclude_wlm=[[0, 0]], order=7,
                    weight_fit_median=0.5, plot=False, verbose=False, fig_size=12):
    """
    THIS IS NOT EXACTLY THE SAME THING THAT applying signal.medfilter()

    This needs to be checked, updated, and combine (if needed) with task fit_smooth_spectrum.
    The task gets the median value in steps of "step", gets an interpolated spectrum, 
    and fits a 7-order polynomy.

    It returns fit_median + fit_median_interpolated (each multiplied by their weights).

    Tasks that use this:  get_telluric_correction
    """

    if verbose:
        vprint("\n> Computing smooth spectrum...")

    if wave_min == 0:
        wave_min = wlm[0]
    if wave_max == 0:
        wave_max = wlm[-1]

    running_wave = []
    running_step_median = []
    cuts = np.int((wave_max - wave_min) / step)

    exclude = 0
    corte_index = -1
    for corte in range(cuts+1):
        next_wave = wave_min+step*corte
        if next_wave < wave_max:
            if next_wave > exclude_wlm[exclude][0] and next_wave < exclude_wlm[exclude][1]:
                if verbose:
                    vprint(f"  Skipping {next_wave}"
                           + f" as it is in the exclusion range [{exclude_wlm[exclude][0]}, {exclude_wlm[exclude][1]}]")

            else:
                corte_index = corte_index+1
                running_wave.append(next_wave)
                region = np.where(
                    (wlm > running_wave[corte_index]-step/2) & (wlm < running_wave[corte_index]+step/2))
                running_step_median.append(np.nanmedian(s[region]))
                if next_wave > exclude_wlm[exclude][1]:
                    exclude = exclude + 1
                    # if verbose and exclude_wlm[0] != [0,0] : print "--- End exclusion range ",exclude
                    if exclude == len(exclude_wlm):
                        exclude = len(exclude_wlm)-1

    running_wave.append(wave_max)
    region = np.where((wlm > wave_max-step) & (wlm < wave_max+0.1))
    running_step_median.append(np.nanmedian(s[region]))

    # Check not nan
    _running_wave_ = []
    _running_step_median_ = []
    for i in range(len(running_wave)):
        if np.isnan(running_step_median[i]):
            if verbose:
                vprint(f"There is a nan in {running_wave[i]}")
        else:
            _running_wave_.append(running_wave[i])
            _running_step_median_.append(running_step_median[i])

    fit = np.polyfit(_running_wave_, _running_step_median_, order)
    pfit = np.poly1d(fit)
    fit_median = pfit(wlm)

    interpolated_continuum_smooth = interpolate.splrep(
        _running_wave_, _running_step_median_, s=0.02)
    fit_median_interpolated = interpolate.splev(
        wlm, interpolated_continuum_smooth, der=0)

    if plot:
        plt.figure(figsize=(fig_size, fig_size/2.5))
        plt.plot(wlm, s, alpha=0.5)
        plt.plot(running_wave, running_step_median, "+", ms=15, mew=3)
        plt.plot(wlm, fit_median, label="fit median")
        plt.plot(wlm, fit_median_interpolated, label="fit median_interp")
        plt.plot(wlm, weight_fit_median*fit_median + (1-weight_fit_median)
                 * fit_median_interpolated, label="weighted")
        # extra_display = (np.nanmax(fit_median)-np.nanmin(fit_median)) / 10
        # plt.ylim(np.nanmin(fit_median)-extra_display, np.nanmax(fit_median)+extra_display)
        ymin = np.nanpercentile(s, 1)
        ymax = np.nanpercentile(s, 99)
        rango = (ymax-ymin)
        ymin = ymin - rango/10.
        ymax = ymax + rango/10.
        plt.ylim(ymin, ymax)
        plt.xlim(wlm[0]-10, wlm[-1]+10)
        plt.minorticks_on()
        plt.legend(frameon=False, loc=1, ncol=1)

        plt.axvline(x=wave_min, color='k', linestyle='--')
        plt.axvline(x=wave_max, color='k', linestyle='--')

        plt.xlabel(r"Wavelength [$\mathrm{\AA}$]")

        if exclude_wlm[0][0] != 0:
            for i in range(len(exclude_wlm)):
                plt.axvspan(exclude_wlm[i][0],
                            exclude_wlm[i][1], color='r', alpha=0.1)
        plt.close()
        vprint(f"Weights for getting smooth spectrum:\n fit_median ={weight_fit_median}"
               + f"\n Fit_median_interpolated = {(1-weight_fit_median)}")

    # (fit_median+fit_median_interpolated)/2      # Decide if fit_median or fit_median_interpolated
    return weight_fit_median*fit_median + (1-weight_fit_median)*fit_median_interpolated

def pixel_in_square(pixel_pos, pixel_size, pos, radius):
    """Compute the area of a pixel within a square.

    Parameters
    ----------
    - pixel_pos: tuple
        Position of the lower left corner of the pixel
    - pixel_size: float
        Size of the pixel.
    - pos: tuple
        Position of the square centre
    - radius: float
        Half size of the square.

    Returns
    -------
    - area_pixel:
        Area of the pixel contained within the square
    - area_fraction:
        Fration of the square area that overlaps with the pixel.
    """
    square = geometry.box(pos[0] - radius,
                          pos[1] - radius,
                          pos[0] + radius,
                          pos[1] + radius)
    rectangle = geometry.box(pixel_pos[0], pixel_pos[1],
                             pixel_pos[0] + pixel_size,
                             pixel_pos[1] + pixel_size)    
    intersection = square.intersection(rectangle)
    return intersection.area, intersection.area / square.area

def pixel_in_circle(pixel_pos, pixel_size, pos, radius):
    """Compute the area of a pixel within a circle.

    Parameters
    ----------
    - pixel_pos: tuple
        Position of the lower left corner of the pixel
    - pixel_size: float
        Size of the pixel.
    - pos: tuple
        Position of the circle centre
    - radius: float
        Radius of the circle.

    Returns
    -------
    - area_pixel:
        Area of the pixel contained within the circle
    - area_fraction:
        Fration of the circle area that overlaps with the pixel.
    """
    circle = geometry.Point(*pos).buffer(radius)
    rectangle = geometry.box(pixel_pos[0], pixel_pos[1],
                             pixel_pos[0] + pixel_size,
                             pixel_pos[1] + pixel_size)    
    intersection = circle.intersection(rectangle)
    return intersection.area, intersection.area / circle.area

def pixel_in_hexagon(pixel_pos, pixel_size, pos, radius):
    """Copmute the area of a pixel overlapping with a regular hexagon.
    
    Description
    -----------
    The hexagon is assumed to be regular.

    Parameters
    ----------
    - pixel_pos: tuple
        Position of the lower left corner of the pixel
    - pixel_size: float
        Size of the pixel.
    - pos: tuple
        Position of the hexagon centre.
    - radius: float
        Radius of the circle that inscribes the hexagon. In other words, the
        distance of the hexagon vertices to the centre.

    Returns
    -------
    - area_pixel:
        Area of the pixel contained within the circle
    - area_fraction:
        Fration of the circle area that overlaps with the pixel.
    """
    # cos(30) * rad / sin(30) * rad
    rad_cos_30 = 0.866 * radius
    rad_sin_30 = 0.5
    # Coordinates are provided in anticlock-wise order starting from the top
    hexagon = geometry.Polygon([
        (pos[0], pos[1] + radius),
        (pos[0] - rad_cos_30, pos[1] + rad_sin_30),
        (pos[0] - rad_cos_30, pos[1] - rad_sin_30),
        (pos[0], pos[1] - radius),
        (pos[0] + rad_cos_30, pos[1] - rad_sin_30),
        (pos[0] + rad_cos_30, pos[1] + rad_sin_30),
        (pos[0], pos[1] + radius)])
    rectangle = geometry.box(pixel_pos[0], pixel_pos[1],
                             pixel_pos[0] + pixel_size,
                             pixel_pos[1] + pixel_size)    
    intersection = hexagon.intersection(rectangle)
    return intersection.area, intersection.area / hexagon.area

# ----------------------------------------------------------------------------------------------------------------------
# Models and fitting
# ----------------------------------------------------------------------------------------------------------------------

def cumulative_1d_moffat(r2, l_star=1.0, alpha2=1.0, beta=1.0):
    """
    Cumulative Moffat ligth profile.
    Parameters
    ----------
    r2 : np.array(float)
        Square radius with respect to the profile centre.
    l_star : float
        Total luminosity integrating from 0 to inf.
    alpha2 : float
        Characteristic square radius.
    beta : float
        Power-low slope
    Returns
    -------
    cum_moffat_prof: np.array(float)
        Cumulative Moffat profile
    """
    return l_star * (1 - np.power(1 + (r2 / alpha2), -beta))

# =============================================================================
# Lines
# =============================================================================
# TODO : merge/remove with future "spectra" module

lines = {
    # Balmer
    'hepsilon': 3970.1,
    'hdelta': 4101.7,
    'hgamma': 4340.4,
    'hbeta': 4861.3,
    'halpha': 6562.79,
    # K
    'K': 3934.777,
    'H': 3969.588,
    'Mg': 5176.7,
    'Na': 5895.6,
    'CaII1': 8500.36,
    'CaII2': 8544.44,
    'CaII3': 8664.52,
}


def mask_lines(wave_array, width=30, lines=lines.values()):
    wave_array = check_unit(wave_array, u.AA)
    width = check_unit(width, wave_array.unit)
    if width.size == 1:
        width = width * np.ones(len(lines))
    mask = np.ones(wave_array.size, dtype=bool)
    for line, w in zip(lines, width):
        line = check_unit(line, u.AA)
        mask[(wave_array < line + w) & (wave_array > line - w)] = False
    return mask

def mask_telluric_lines(wave_array, width_clip=3 << u.AA):
    model_file = os.path.join(os.path.dirname(__file__),
                                      'input_data', 'sky_lines',
                                      'telluric_lines.txt')
    w_l_1, w_l_2 = np.loadtxt(model_file, unpack=True, usecols=(0, 1))
    w_l_1 = w_l_1 << u.angstrom
    w_l_2 = w_l_2 << u.angstrom
    return mask_lines(wave_array, width= np.clip(w_l_2 - w_l_1, a_min=width_clip, a_max=None),
                      lines=(w_l_1 + w_l_2) / 2)

# =============================================================================
# Mr Krtxo \(ﾟ▽ﾟ)/
#                                                       ... Paranoy@ Rulz! ;^D



"""
This module contains tools for creating synthetic photometry from DataContainers
as well as tools for retrieveing and manipulating external imaging data.

""" 
import os
import requests
from io import StringIO

import numpy as np
import matplotlib.pyplot as plt

from astropy import units as u
from astropy.table import Table
from astropy.nddata import CCDData, StdDevUncertainty
from astropy.coordinates import SkyCoord
from photutils.aperture import SkyCircularAperture, ApertureStats

from scipy.optimize import minimize
from scipy.ndimage import gaussian_filter

from pykoala import vprint
from pykoala.data_container import Cube, RSS
from pykoala.ancillary import update_wcs_coords
from pykoala.plotting import utils

class QueryMixin:
    """Mixin with common methods for image queries to external databases."""

    def download_image(url, filename):
        """Download an image and write it to a binary file.
        
        Parameters
        ----------
        url : str
            URL to download the image.
        filename : str
            Name of the output binary file.
        
        Returns
        -------
        filename : str
            Output filename. If the download is unsuccessful, the return value
            is None.
        """
        vprint(f"Downloading: {url}")
        try:
            r = requests.get(url)
        except Exception as e:
            vprint(f"ERROR: Download unsuccessful (error: {e})")
            return None
        vprint(f"Saving file at: {filename}")
        with open(filename,"wb") as f:
            f.write(r.content)
        return filename

    def filename_from_pos(ra, dec, filter, survey):
        """Convenience function for creating a filename based on the query information.
        
        This method creates a filename using the central (RA, DEC) values of the
        querie, the filter name associated to the queried image, and the
        survey/database name. The resulting filename uses the following convention:
        ``SURVEY_query_RA_DEC_FILTER.fits``, where negative values of DEC include
        the prefix ``n`` (i.e. `n30` for ``dec=-30``)

        Parameters
        ----------
        ra : float
            Reference RA position in deg.
        dec : float
            Reference DEC position in deg.
        filter : str
            Photometric filter name.
        survey : str
            Survey name.
        
        Returns
        -------
        filename : str
            Output filename
        """
        sign = np.sign(dec)
        if sign == -1:
            sign_str = 'n'
        else:
            sign_str = ''
        filename = f"{survey}_query_{ra:.4f}_{sign_str}{np.abs(dec):.4f}_{filter}.fits"
        return filename


class LegacySurveyQuery(QueryMixin):
    """Utility to query `LegacySurvey <https://www.legacysurvey.org>`_ imaging data.
    
    Attributes
    ----------
    fitscutout : str
        URL to the online cutout service.
    dafault_layer : str, default=``ls-dr10``
        LS layer used to perform the query.
    pixelsize_arcsec : size of the pixels in angstrom.
    """
    fitscutout = "https://www.legacysurvey.org/viewer/fits-cutout"
    default_layer = "ls-dr10"
    pixelsize_arcsec = 0.27 << u.arcsec

    @classmethod
    def getimage(cls, im_coords, size=240, filters="r", images={}, output_dir="."):
        ra, dec = im_coords.ra.to_value("deg"), im_coords.dec.to_value("deg")
        for f in filters:
            url = cls.fitscutout
            url += f"?ra={ra:.5f}&dec={dec:.5f}&layer={cls.default_layer}"
            url += f"&pixscale={cls.pixelsize_arcsec}&bands={f}&size={size}"
            filename = cls.filename_from_pos(ra, dec, f, "ls")
            output = os.path.join(output_dir, filename)
            filename = cls.download_image(url, filename=output)


            if filename is not None:
                ccd = cls.read_image(filename)
                images[f"LS.{f}"] = {"ccd": ccd,
                                     "pix_size": cls.pixelsize_arcsec}

    def read_image(filename):
        """Read a LS fits image.
        
        Parameters
        ----------
        filename : str
            Path to the FITS file containing the image data.

        Returns
        -------
        ccd : ``astropy.nddata.CCDData``
            LS image data. The default data units are Jy.
        """
        vprint("Opening Legacy Survey fits file")
        ccd = CCDData.read(filename, hdu=0, unit="adu")
        # Convert ADU to Jy
        ccd = ccd.multiply(3631e-9 << u.Jy / u.adu)
        return ccd


class PSQuery(QueryMixin):
    """Query to the PANSTARRS (DR2) Survey.
    
    Attributes
    ----------
    ps1filename : str
        URL to the PS query service
    fitscut : str
        URL to the cutour service
    pixelsize_arcsec : size of the pixels in angstrom.
    """

    ps1filename = "https://ps1images.stsci.edu/cgi-bin/ps1filenames.py"
    fitscut = "https://ps1images.stsci.edu/cgi-bin/fitscut.cgi"
    pixelsize_arcsec = 0.25 << u.arcsec

    def getimages(tra, tdec, size=240, filters="grizy", format="fits",
                  imagetypes="stack"):    
        """

        Description
        -----------
        Code selected from:
        https://outerspace.stsci.edu/display/PANSTARRS/PS1+Image+Cutout+Service#PS1ImageCutoutService-python

        Query ps1filenames.py service for multiple positions to get a list of images
        This adds a url column to the table to retrieve the cutout.
        
        tra, tdec = list of positions in degrees
        size = image size in pixels (0.25 arcsec/pixel)
        filters = string with filters to include
        format = data format (options are "fits", "jpg", or "png")
        imagetypes = list of any of the acceptable image types.  Default is stack;
            other common choices include warp (single-epoch images), stack.wt (weight image),
            stack.mask, stack.exp (exposure time), stack.num (number of exposures),
            warp.wt, and warp.mask.  This parameter can be a list of strings or a
            comma-separated string.
    
        Returns an astropy table with the results
        """
        
        if format not in ("jpg","png","fits"):
            raise ValueError("format must be one of jpg, png, fits")
        # if imagetypes is a list, convert to a comma-separated string
        if not isinstance(imagetypes,str):
            imagetypes = ",".join(imagetypes)
        # put the positions in an in-memory file object
        cbuf = StringIO()
        cbuf.write('\n'.join(["{} {}".format(ra, dec) for (ra, dec) in zip(tra,tdec)]))
        cbuf.seek(0)
        # use requests.post to pass in positions as a file
        r = requests.post(PSQuery.ps1filename, data=dict(filters=filters, type=imagetypes),
            files=dict(file=cbuf))
        r.raise_for_status()
        tab = Table.read(r.text, format="ascii")
    
        urlbase = "{}?size={}&format={}".format(PSQuery.fitscut,size,format)
        tab["url"] = ["{}&ra={}&dec={}&red={}".format(urlbase,ra,dec,filename)
                for (filename,ra,dec) in zip(tab["filename"],tab["ra"],tab["dec"])]
        return tab

    @classmethod
    def getimage(cls, im_coords, size=240, filters="grizy", format="fits",
                 imagetypes="stack", images={}, output_dir="."):
        
        """Query ps1filenames.py service for multiple positions to get a list of images
        This adds a url column to the table to retrieve the cutout.
        
        tra, tdec = list of positions in degrees
        size = image size in pixels (0.25 arcsec/pixel)
        filters = string with filters to include
        format = data format (options are "fits", "jpg", or "png")
        imagetypes = list of any of the acceptable image types.  Default is stack;
            other common choices include warp (single-epoch images), stack.wt (weight image),
            stack.mask, stack.exp (exposure time), stack.num (number of exposures),
            warp.wt, and warp.mask.  This parameter can be a list of strings or a
            comma-separated string.
    
        Returns an astropy table with the results
        """
        
        if format not in ("jpg","png","fits"):
            raise ValueError("format must be one of jpg, png, fits")
        # if imagetypes is a list, convert to a comma-separated string
        if not isinstance(imagetypes,str):
            imagetypes = ",".join(imagetypes)
        # put the positions in an in-memory file object
        cbuf = StringIO()
        cbuf.write('\n'.join(["{} {}".format(im_coords.ra.to_value("deg"),
                                             im_coords.dec.to_value("deg"))]))
        cbuf.seek(0)
        # use requests.post to pass in positions as a file
        r = requests.post(cls.ps1filename, data=dict(filters=filters, type=imagetypes),
            files=dict(file=cbuf))
        r.raise_for_status()
        tab = Table.read(r.text, format="ascii")
    
        urlbase = "{}?size={}&format={}".format(cls.fitscut,size,format)
        tab["url"] = ["{}&ra={}&dec={}&red={}".format(urlbase,ra,dec,filename)
                for (filename,ra,dec) in zip(tab["filename"],tab["ra"],tab["dec"])]
        # Retrieve each filter image
        for row in tab:
            vprint(f"Retrieving cutout: ra={row['ra']}, dec={row['dec']}, filter={row['filter']}")
            sign = np.sign(row['dec'])
            if sign == -1:
                sign_str = 'n'
            else:
                sign_str = ''
            filename = f"ps_query_{row['ra']:.4f}_{sign_str}{np.abs(row['dec']):.4f}_{row['filter']}.fits"
            filename.replace("-", "n")
            output = os.path.join(output_dir, filename)
            filename = cls.download_image(row['url'], filename=output)
            if filename is not None:
                ccd = cls.read_ps_fits(filename)
                images[f"PS1.{row['filter']}"] = {
                        "ccd": ccd, "pix_size": cls.pixelsize_arcsec}

        return images

    def read_ps_fits(filename):
        """Load a PANSTARRS image.
        
        Parameters
        ----------
        filename : str
            Path to the FITS file containing the image data.

        Returns
        -------
        ccd : ``astropy.nddata.CCDData``
            PS image data. The default data units are Jy.
        """
        vprint("Opening PANSTARRS fits file")
        ccd = CCDData.read(filename, hdu=0, unit="adu")
        # Convert ADU to Jy
        ccd = ccd.multiply(3631 * 10**(-0.4 * ccd.header["FPA.ZP"]
                                ) / ccd.header["exptime"] << u.Jy / u.adu)
        return ccd


def get_effective_sky_footprint(data_containers):
    """Compute the effective footprint containing all input DataContainers.
    
    Returns
    -------
    centre: tuple
        Centre coordinates (ra_mean, dec_mean)
    width: tuple
        Window width (ra_max - ra_min, dec_max - dec_min)
    """
    data_containers_footprint = []
    for dc in data_containers:
        data_containers_footprint.append(dc.get_footprint())
    # Select a rectangle containing all footprints
    # TODO: Implement units on DC and remove << deg
    max_ra, max_dec = np.nanmax(data_containers_footprint, axis=(0, 1)) << u.deg
    min_ra, min_dec = np.nanmin(data_containers_footprint, axis=(0, 1)) << u.deg
    centre_coordinates = SkyCoord((max_ra + min_ra) / 2, (max_dec + min_dec) / 2)
    ra_width, dec_width = max_ra - min_ra, max_dec - min_dec
    vprint("Combined footprint Fov: {}, {}".format(ra_width.to("arcmin"),
                                                   dec_width.to("arcmin")))
    return centre_coordinates, (ra_width, dec_width)

def query_image(data_containers, query=PSQuery, filters='r',
                im_extra_size_arcsec=30 << u.arcsec, im_output_dir='.'):
    """Perform a query of external images that overlap with the DataContainers.
    
    This method performs a query to the database of some photometric survey (e.g. PS)
    and retrieves a set of images that overlap with the IFS data.

    Parameters
    ----------
    survey : str, default="PS"
        Name of the external survey/database to perform the query. At present only PS
        queries are available.
    filters : str, default='r'
        String containing the filters to be included in the query (e.g. "ugriz").
    im_extra_size_arcsec : float, default=30
        Additional extension of the images in arcseconds with respect to the net FoV
        of the DataContainers.
    im_output_dir : str, default='.'
        Path to a directory where the queried images will be stored. Default is current
        working directory.

    Returns
    -------
    images : dict
        Dictionary containing the image data (``intensity``), wcs (``wcs``)
        and pixel size (``pix_size``) of each queried filter.
    """
    vprint("Querying image to external database")

    # Compute the effective footprint of all DC and use that as input for
    # the query
    im_coords, im_fov = get_effective_sky_footprint(data_containers)
    # Convert the size to pixels
    ra_pixels = ((im_fov[0] + im_extra_size_arcsec) / query.pixelsize_arcsec
                 ).to_value(u.dimensionless_unscaled)
    dec_pixels = ((im_fov[1] + im_extra_size_arcsec) / query.pixelsize_arcsec
                  ).to_value(u.dimensionless_unscaled)

    im_size_pix = int(np.max([ra_pixels, dec_pixels]))
    vprint(f"Image center sky position (RA, DEC): {im_coords}")
    vprint(f"Image size (pixels): {im_size_pix}")
    # Perform the query
    images = query.getimage(im_coords, size=im_size_pix, filters=filters,
                            output_dir=im_output_dir)
    return images

def get_aperture_photometry(coordinates : SkyCoord, diameters : u.Quantity,
                            image : CCDData):
    """Compute the aperture photometry from an input image.
    
    Convenient method to compute aperture photometry using multiple circular
    apertures. The flux within each aperture is computed as the mean value
    multiplied by the exact number of pixels contained within.

    Parameters
    ----------
    coordinates : :class:`astropy.coordinates.SkyCoord`
        Celestial coordinates where to compute the apertures.
    diameters : :class:`astropy.units.Quantity`
        Diameter size of each aperture. It can be a single or
        multiple values.
    image : :class:`astropy.nddata.CCDData`

    Returns
    -------
    flux_in_ap : :class:`astropy.units.Quantity`
        Flux contained within each aperture.
    flux_in_ap_err : :class:`astropy.units.Quantity`
        Associated error to ``flux_in_ap``.
    """
    apertures = SkyCircularAperture(coordinates, r=diameters / 2)
    table = ApertureStats(image, apertures, sum_method='exact')
    flux_in_ap = table.mean * np.sqrt(table.center_aper_area.value)
    flux_in_ap_err = table.sum_err
    return flux_in_ap, flux_in_ap_err


def get_dc_aperture_flux(data_container, filter_name,
                         aperture_diameter=1.25 << u.arcsec,
                         sample_every=2, rss_fibres_pct=50.0):
    """Compute aperture fluxes from the DataContainers

    This method computes a set of aperture fluxes from an input data container.
    If the input DC is a Cube, a grid of equally-spaced apertures will be
    computed. If the input DC is a RSS, the fibre positions will be used as
    reference apertures.

    Parameters
    ----------
    data_container : :class:`pykoala.data_container.SpectraContainer`
        DataContainer used to compute the synthetic photometry.
    filter_name: str
        Photometric filter name used to initalise a :class:`pst.observables.Filter`.
    aperture_diameter: float
        Diameter size of the circular apertures. In the case of an RSS, this
        will match the size of the fibres.
    sample_every: int, default=2
        Spatial aperture sampling in units of the aperture radius. If
        `sample_every=2`, the aperture will be defined every two aperture
        diameters in the image.
    rss_fibres_pct : float, optional. Default=50
        If the input DataContainer is :class:`RSS`, it determines the
        flux percentile threshold used to select valid fibre apertures.

    Returns
    -------
    dc_photometry: dict
        A dictionary containing the results of the computation

        - synth_photo: The synthetic photometry of the DC, it can be a list
        of fluxes (if the input DC is a RSS) or an image (if DC is a Cube).
        For more details, see the method `get_dc_synthetic_photometry`.
        - synth_photo_err: Associated error fo `synth_photo`.
        - wcs: WCS of the DC. If the DC is a RSS it will be None.
        - coordinates: Celestial position of the apertures stored as a list
        of `astropy.coordinates.Skycoord`.
        - aperture_mask: A mask for invalid aperture fluxes.
        - figs: A list of QC figures showing the .
    
    See also
    --------
    :class:`pst.observables.Filter`
    """
    try:
        from pst.observables import Filter
    except:
        raise ImportError("PST package not found")

    result = {}

    # Compute the synthetic photometry maps    
    photometric_filter = Filter.from_svo(filter_name)
    synth_photo, synth_photo_err = get_dc_synthetic_photometry(
        photometric_filter, data_container)
    result['synth_photo'] = synth_photo
    result['synth_photo_err'] = synth_photo_err

    if isinstance(data_container, Cube):
        vprint("Computing aperture fluxes using Cube synthetic photometry")
        # Create a grid of apertures equally spaced
        pix_size = np.max(
            data_container.wcs.celestial.wcs.cdelt) * 3600 * u.arcsec
        delta_pix = aperture_diameter / pix_size * sample_every
        vprint("Creating a grid of circular aperture "
                + f"(rad={aperture_diameter / 2 / pix_size:.2f}"
                + f" px) every {delta_pix:.1f} pixels")
        rows = np.arange(0, synth_photo.shape[0], delta_pix)
        columns = np.arange(0, synth_photo.shape[1], delta_pix)
        yy, xx = np.meshgrid(rows, columns)
        coordinates = data_container.wcs.celestial.pixel_to_world(
            xx.flatten(), yy.flatten())
        
        flux_in_ap, flux_in_ap_err = get_aperture_photometry(
            coordinates, diameters=aperture_diameter,
            image=CCDData(data=synth_photo, uncertainty=StdDevUncertainty(
                synth_photo_err),
            wcs=data_container.wcs.celestial))
        result['wcs'] = data_container.wcs.celestial.deepcopy()

    elif isinstance(data_container, RSS):
        vprint("Using RSS fibre synthetic photometry as apertures")
        mask = synth_photo > np.nanpercentile(synth_photo, rss_fibres_pct)
        # TODO: this should be a method or an attribute of RSS.
        coordinates = SkyCoord(data_container.info['fib_ra'][mask],
                               data_container.info['fib_dec'][mask],
                               unit='deg')
        flux_in_ap, flux_in_ap_err = synth_photo[mask], synth_photo_err[mask]
        result['wcs'] = None
        aperture_diameter = data_container.fibre_diameter

    # Make a QC plot of the apertures
    fig = make_plot_apertures(
        data_container, synth_photo, synth_photo_err, coordinates, flux_in_ap,
        flux_in_ap_err)
    # Store the results
    result['fig'] = fig
    result['coordinates'] = coordinates
    result['aperture_flux'] = flux_in_ap
    result['aperture_flux_err'] = flux_in_ap_err
    result['aperture_mask'] =  np.isfinite(flux_in_ap
                                           ) & np.isfinite(flux_in_ap_err)
    return result

def make_plot_apertures(dc, synth_phot, synth_phot_err, ap_coords,
                            ap_flux, ap_flux_err):
    """Plot the synthetic aperture fluxes measured from a DC."""
    if isinstance(dc, Cube):
        fig, axs = plt.subplots(ncols=2, nrows=2, sharex=True, sharey=True,
                                subplot_kw={'projection': dc.wcs.celestial},
                                constrained_layout=True)
        ax = axs[0, 0]
        mappable = ax.imshow(-2.5 * np.log10(synth_phot.to_value("3631 Jy")), vmin=16, vmax=23,
                                interpolation='none')
        plt.colorbar(mappable, ax=ax, label='SB (mag/pix)')
        ax = axs[0, 1]
        mappable = ax.imshow(synth_phot / synth_phot_err, vmin=0, vmax=10, cmap='jet',
                                interpolation='none')
        plt.colorbar(mappable, ax=ax, label='Flux SNR')
        ax = axs[1, 0]            
        mappable = ax.scatter(ap_coords.ra, ap_coords.dec,
        marker='o', transform=ax.get_transform('world'),
        c= -2.5 * np.log10(ap_flux.to_value("3631 Jy")), vmin=16, vmax=23)
        plt.colorbar(mappable, ax=ax, label='SB (mag/aperture)')
        ax = axs[1, 1]
        mappable = ax.scatter(ap_coords.ra, ap_coords.dec,
        marker='o', transform=ax.get_transform('world'),
        c= ap_flux /ap_flux_err, vmin=0, vmax=10)
        plt.colorbar(mappable, ax=ax, label='Aper flux SNR')
        for ax in axs.flatten():
            ax.coords.grid(True, color='orange', ls='solid')
            ax.coords[0].set_format_unit('deg')
        plt.close(fig)
    elif isinstance(dc, RSS):
        fig, axs = plt.subplots(ncols=1, nrows=2, sharex=True, sharey=True,
                                constrained_layout=True)
        mappable = axs[0].scatter(dc.info['fib_ra'], dc.info['fib_dec'],
                                    c=synth_phot)
        plt.colorbar(mappable, label="Flux")
        mappable = axs[1].scatter(dc.info['fib_ra'], dc.info['fib_dec'],
                                    c=synth_phot / synth_phot_err, label="SNR")
        plt.colorbar(mappable)
        # TODO: once plotting is homogeneized add RSS fibre map
    return fig

def get_dc_synthetic_photometry(filter, dc):
    """Compute synthetic photometry from a DataContainer.
    
    This method extracts synthetic photometry using the spectral information
    of a DataContainer.

    Parameters
    ----------
    filter: :class:`pst.observables.Filter`
        A Filter for computing the synthetic photometry from the spectra.
    dc: :class:`DataContainer`
        The input DataContainer.
    dc_intensity_units: :class:`astropy.units.Quantity`
        The units of the intensity of the DC.

    Returns
    -------
    - synth_phot: :class:`np.ndarray`
        Array containing the flux estimates expressed in Jy.
    - synth_phot_err: :class:`np.ndarray`
        Array containing the error associated to the flux estimate.
    """
    # interpolate the filter response curve to the DC wavelength grid.
    filter.interpolate(dc.wavelength)
    synth_photo = np.full((dc.rss_intensity.shape[0], 1), fill_value=np.nan) << u.Jy
    synth_photo_err = np.full_like(synth_photo, fill_value=np.nan)

    for ith, (intensity, variance) in enumerate(
        zip(dc.rss_intensity, dc.rss_variance)):
        mask = np.isfinite(intensity) & np.isfinite(variance)
        if not mask.any():
                continue
        #TODO: Temporarily until units are implemented in DC.
        if not isinstance(intensity, u.Quantity):
            intensity = intensity * 1e-16 * u.erg / u.s / u.cm**2 / u.angstrom
            variance = variance * (1e-16 * u.erg / u.s / u.cm**2 / u.angstrom)**2
        f_nu, f_nu_err = filter.get_fnu(intensity, variance**0.5)
        synth_photo[ith] = f_nu.to('Jy')
        synth_photo_err[ith] = f_nu_err.to('Jy')

    synth_photo = np.squeeze(dc.rss_to_original(synth_photo))
    synth_photo_err = np.squeeze(dc.rss_to_original(synth_photo_err))

    return synth_photo, synth_photo_err

# def crosscorrelate_im_apertures(ref_aperture_flux, ref_aperture_flux_err,
#                                 ref_coord, image,
#                                 ra_offset_range=[-10, 10],
#                                 dec_offset_range=[-10, 10],
#                                 offset_step=0.5,
#                                 aperture_diameter=1.25 << u.arcsec,
#                                 plot=True):
#     """Cross-correlate an image with an input set of apertures.
    
#     Description
#     -----------
#     This method performs a spatial cross-correlation between a list of input aperture fluxes
#     and a reference image. For example, the aperture fluxes can simply correspond to the
#     flux measured within a fibre or a set of aperture measured in a datacube.
#     First, a grid of aperture position offsets is generated, and every iteration
#     will compute the difference between the two sets. 

#     The figure of merit used is defined as:
#         :math: `w=e^{-(A+B)/2}`
#         where 
#         :math: `A=\langle f_{Ap} - \hat{f_{Ap}} \rangle`
#         :math: `B=|1 - \frac{\langle f_{Ap} \rangle}{\langle \hat{f_{Ap}} \rangle}|`
#         where :math: `f_{Ap}, \hat{f_{Ap}` correspond to the aperture flux in the
#         reference (DC) and ancillary data

#     Parameters
#     ----------
#     ref_aperture_flux: np.ndarray
#         A set of aperture fluxes measured in the target image.
#     ref_aperture_flux_err: np.ndarray
#         The associated errors of the aperture fluxes.
#     ref_coords: astropy.coordinates.SkyCoord
#         The celestial coordinates of each aperture.
#     image: np.ndarray
#         Pixel data associated to the reference image.
#     wcs: astropy.wcs.WCS
#         WCS associated to the reference image.
#     ra_offset_range: list or tuple, default=[-10, 10],
#         The range of offsets in RA to be explored in arcseconds.
#         Defaul is +-10 arcsec.
#     dec_offset_range: list or tuple, default=[-10, 10],
#         The range of offsets in DEC to be explored in arcseconds.
#         Defaul is +-10 arcsec.
#     offset_step: float, default=0.5
#         Offset step size in arcseconds. Default is 0.5.
#     aperture_diameter: float, default=1.25
#         Aperture diameter size in arcseconds. Default is 1.25.

#     Returns
#     -------
#     results: dict
#         Dictionary containing the cross-correlation results.
#     """
#     vprint("Cross-correlating image to list of apertures")
#     vprint(f"Input number of apertures: {len(ref_aperture_flux)}")
#     # Renormalize the reference aperture
#     ref_ap_norm = np.nanmean(ref_aperture_flux)
#     ref_flux = ref_aperture_flux / ref_ap_norm
#     ref_flux_err = ref_aperture_flux_err / ref_ap_norm
#     good_aper = np.isfinite(ref_flux) & (ref_flux > 0)
#     vprint(f"Number of input apertures used: {np.count_nonzero(good_aper)}")
#     # Make a grid of offsets
#     ra_offset = np.arange(*ra_offset_range, offset_step)
#     dec_offset = np.arange(*dec_offset_range, offset_step)

#     # Initialise the results variable

#     offset_sampling = np.meshgrid(dec_offset, ra_offset, indexing='ij')
#     grid_flux_prod = np.full((dec_offset.size, ra_offset.size),
#                               fill_value=np.nan)

#     for i, (ra_offset_arcsec, dec_offset_arcsec) in enumerate(
#         zip(offset_sampling[0].flatten(), offset_sampling[1].flatten())):
#         # Create a set of apertures
#         new_coords = SkyCoord(
#             ref_coord.ra + ra_offset_arcsec * u.arcsec,
#             ref_coord.dec + dec_offset_arcsec * u.arcsec)

#         flux_in_ap, flux_in_ap_err = get_aperture_photometry(
#         new_coords, diameters=aperture_diameter, image=image["ccd"])

#         # Renormalize the flux
#         flux_in_ap_norm = np.nanmean(flux_in_ap)
#         flux_in_ap /= flux_in_ap_norm
#         flux_in_ap_err /= flux_in_ap_norm
#         # Ensure that none of the apertures contain NaN's/inf
#         mask = np.isfinite(flux_in_ap) & good_aper
#         n_valid = np.count_nonzero(mask)
#         idx = np.unravel_index(i, grid_flux_prod.shape)
#         grid_flux_prod[idx] = np.nansum(
#             flux_in_ap[mask] * ref_flux[mask]) / n_valid

#     vprint("Computing the offset solution")
#     weights = np.nanmax(grid_flux_prod) - grid_flux_prod
#     weights = np.exp(grid_flux_prod)
#     weights /= np.nansum(weights)
#     # Minimum
#     min_pos = np.nanargmax(weights)
#     min_pos = np.unravel_index(min_pos, grid_flux_prod.shape)

#     ra_min = offset_sampling[1][min_pos]
#     dec_min = offset_sampling[0][min_pos]
#     ra_mean = np.nansum(offset_sampling[1] * weights)
#     ra_var = np.nansum((offset_sampling[1] - ra_mean)**2 * weights)
#     dec_mean = np.nansum(offset_sampling[0] * weights)
#     dec_var = np.nansum((offset_sampling[0] - dec_mean)**2 * weights)

#     results = {
#         "offset_min": (ra_min, dec_min),
#         "offset_mean": (ra_mean, dec_mean),
#         "offset_var": (ra_var, dec_var),
#         "ra_offset": ra_offset, "dec_offset": dec_offset,
#         "offset_sampling": offset_sampling,
#         "weights": weights}

#     if plot:
#         results['fig'] = make_crosscorr_plot(results)
#     return results

# def make_crosscorr_plot(results):
#     """Make a plot showing the aperture flux cross-correlatation results.
    
#     Parameters
#     ----------
#     results: dict
#         Dictionary containing the results returned by `crosscorrelate_im_apertures`.

#     Returns
#     -------
#     fig : :class:`plt.Figure`
#         Figure containing the plot.
#     """
#     ra_mean, dec_mean = results['offset_mean']
#     ra_min, dec_min = results['offset_min']

#     fig, ax = plt.subplots(ncols=1, nrows=1, constrained_layout=True,
#                            figsize=(4, 4),
#                            sharex=False, sharey=False)

#     mappable = ax.pcolormesh(
#         results['ra_offset'], results['dec_offset'], results['weights'],
#         cmap='Spectral')
#     ax.plot(ra_mean, dec_mean, 'o', ms=10, label='Mean', mec='k', mfc='none')
#     ax.plot(ra_min, dec_min, 'k+', label='Max. like')
#     #ax.legend(framealpha=0.1)
#     ax.set_xlabel("RA offset (arcsec)")
#     ax.set_ylabel("DEC offset (arcsec)")
#     plt.colorbar(mappable, ax=ax, label='W',
#                  orientation='horizontal', pad=0.2)
    
#     vax = ax.inset_axes((1.03, 0, .6, 1), sharey=ax)
#     vax.plot(results['weights'].sum(axis=1), results['dec_offset'], 'k')
#     vax.axhline(dec_min, color='b', label='Max. weight')
#     vax.axhline(dec_mean, color='r', label='Avg.')
#     vax.legend()

#     hax = ax.inset_axes((0, 1.03, 1, .6), sharex=ax)
#     hax.plot(results['ra_offset'], results['weights'].sum(axis=0), 'k')
#     hax.axvline(ra_min, color='b')
#     hax.axvline(ra_mean, color='r')

#     plt.close(fig)
#     return fig

def crosscorrelate_im_apertures(ref_aperture_flux, ref_coord, image,
                                ra_offset_range=[-10., 10.],
                                dec_offset_range=[-10., 10.],
                                aperture_diameter=1.25 << u.arcsec,
                                smooth_image_sigma=1 << u.arcsec):
    """Cross-correlate an image with an input set of apertures.
    
    Description
    -----------
    This method performs a spatial cross-correlation between a list of input aperture fluxes
    and a reference image. For example, the aperture fluxes can simply correspond to the
    flux measured within a fibre or a set of aperture measured in a datacube.

    The objective function is defined as:
        :math: `f(\Delta RA, \Delta DEC)= \frac{1}{\sum f_i \hat{f_i}(\Delta RA, \Delta DEC)}`
        where :math: `f_i` and :math: `\hat{f_i}` correspond to the flux measured
        within the reference and image apertures.


    Parameters
    ----------
    ref_aperture_flux: np.ndarray
        A set of aperture fluxes measured in the target image.
    ref_coords: astropy.coordinates.SkyCoord
        The celestial coordinates of each aperture.
    image: np.ndarray
        Pixel data associated to the reference image.
    ra_offset_range: list or tuple, default=[-10, 10],
        The range of offsets in RA to be explored in arcseconds.
        Defaul is +-10 arcsec.
    dec_offset_range: list or tuple, default=[-10, 10],
        The range of offsets in DEC to be explored in arcseconds.
        Defaul is +-10 arcsec.
    aperture_diameter: :class:`astropy.units.Quantity`, default=1.25 arcsec
        Aperture diameter size in arcseconds. Default is 1.25.
    smooth_image_sigma: :class:`astropy.units.Quantity` or None, default=1 arcsec
        Gaussian smoothing standard deviation applied to the reference image before
        the cross-correlation. If ``None``, no smoothing is applied.

    Returns
    -------
    results: dict
        Dictionary containing the cross-correlation results.
    """
    vprint("Cross-correlating image to list of apertures")
    vprint(f"Input number of apertures: {len(ref_aperture_flux)}")
    # Renormalize the reference aperture
    ref_ap_norm = np.nanmean(ref_aperture_flux)
    ref_flux = ref_aperture_flux / ref_ap_norm
    good_aper = np.isfinite(ref_flux) & (ref_flux > 0)
    vprint(f"Number of input apertures used: {np.count_nonzero(good_aper)}")
    # Make a grid of offsets
    if smooth_image_sigma is not None:
        image["ccd"].data = gaussian_filter(
            image["ccd"].data,
            smooth_image_sigma.to_value("arcsec") / image["pix_size"].to_value("arcsec"))

    def objective_function(offsets):
        # Create the new aperture coordinates
        new_coords = SkyCoord(
            ref_coord.ra + offsets[0] * u.arcsec,
            ref_coord.dec + offsets[1] * u.arcsec)

        flux_in_ap, _ = get_aperture_photometry(
        new_coords, diameters=aperture_diameter, image=image["ccd"])

        # Renormalize the flux
        flux_in_ap_norm = np.nanmean(flux_in_ap)
        flux_in_ap /= flux_in_ap_norm
        # Ensure that none of the apertures contain NaN's/inf
        mask = np.isfinite(flux_in_ap) & good_aper
        n_valid = np.count_nonzero(mask)
        value = 1 / (np.nansum(flux_in_ap[mask] * ref_flux[mask]) / n_valid)
        return value

    vprint("Performing minimization...")
    result = minimize(objective_function, x0=[0., 0.],
                      bounds=(ra_offset_range, dec_offset_range),
                      method="Powell",
                      tol=1e-9,
                      )
    vprint(f"Success : {result.success}")
    vprint(f"Status : {result.status}")
    vprint(f"Result : {result.x}")
    vprint(f"Number of evaluations : {result.nfev}")
    results = {"offset_min": result.x << u.arcsec, "opt_results": result}

    return results

def make_plot_astrometry_offset(data_container, dc_synth_photo, image, results):
        """Plot the DC and ancillary data including the astrometry correction.
        
        Parameters
        ----------
        data_container : :class:`DataContainer`
            DataContainer used during the cross-correlation.
        dc_synth_photo : :class:`astropy.units.Quantity`
            DataContainer synthetic photometry.
        image : dict
            Dictionary including the keys ``ccd`` and ``pix_size``, containing
            the :class:`CCDData` and pixel size information of the reference
            image.
        results : dict
            Dictionary containing the results of the cross-correlation.
        
        Returns
        -------
        fig : :class:`plt.Figure`
            Figure containing the plots.
        """

        synt_sb = -2.5 * np.log10(dc_synth_photo.to_value("3631 Jy"))
        im_sb = -2.5 * np.log10(
            image['ccd'].data * image['ccd'].unit.to("3631 Jy")
            / image['pix_size'].to_value("arcsec")**2)

        dc_contour_params = dict(levels=[17, 18, 19, 20, 21, 22], colors='r')
        contour_params = dict(levels=[17, 18, 19, 20, 21, 22], colors='k')

        offset_str = (f"{results['offset_min'][0]:.1f}"
                      + f", {results['offset_min'][1]:.1f}")
        
        fig, axs = utils.new_figure(f"Astrometry Offset ({offset_str})", figsize=(10, 4),
                                    ncols=2, sharex=True, sharey=True,
                                    subplot_kw=dict(projection=image['ccd'].wcs),
                                    gridspec_kw=dict(wspace=0.1),
                                    squeeze=True,
                                    constrained_layout=True,
                                    tweak_axes=False
                                    )
        axs[0].set_title("Original")
        axs[1].set_title("Corrected")

        for ax in axs:
            ax.coords.grid(True, color='orange', ls='solid')
            ax.coords[0].set_format_unit('deg')
            cs = ax.contour(im_sb, **contour_params)

        if isinstance(data_container, Cube):
            cs = axs[0].contour(synt_sb, transform=axs[0].get_transform(
                data_container.wcs.celestial), **dc_contour_params)
            # Compute the correctec WCS
            correct_wcs = update_wcs_coords(wcs=data_container.wcs.celestial,
                                            ra_dec_offset=(
                                                results['offset_min'][0],
                                                results['offset_min'][1]))
            cs = axs[1].contour(synt_sb, transform=axs[1].get_transform(correct_wcs),
                           **dc_contour_params)
        elif isinstance(data_container, RSS):
            # TODO: Temporal until the final rss plotting method is in place
            mappable = axs[0].scatter(data_container.info['fib_ra'],
                                      data_container.info['fib_dec'],
                                      c=synt_sb,
                                      transform=axs[0].get_transform("world"))
            #plt.colorbar(mappable, ax=axs[0])
            mappable = axs[1].scatter(
                data_container.info['fib_ra'] + results['offset_min'][0],
                data_container.info['fib_dec'] + results['offset_min'][1],
                c=synt_sb,
                transform=axs[1].get_transform("world"))
            plt.colorbar(mappable, ax=axs[1], label='Aperture magnitude')
        plt.close(fig)
        return fig



import numpy as np
import matplotlib.pyplot as plt
import matplotlib.colors as colors
from pykoala.plotting.quick_plot import quick_plot #, basic_statistics

# Fuego color map
#from matplotlib import pyplot as plt
#import matplotlib.colors as colors

fuego_color_map = colors.LinearSegmentedColormap.from_list("fuego", 
                                                            ((0.25, 0, 0),  
                                                            (0.5,0,0),    
                                                            (1, 0, 0), 
                                                            (1, 0.5, 0), 
                                                            (1, 0.75, 0), 
                                                            (1, 1, 0), 
                                                            (1, 1, 1)), 
                                                            N=256, gamma=1.0)
fuego_color_map.set_bad('lightgray')  #('black')
#plt.register_cmap(cmap=fuego_color_map)  
# #-----------------------------------------------------------------------------
# %% ============================================================================
# #-----------------------------------------------------------------------------

def plot_map():
    """
    Task for plotting a map from a cube. Done in Angel's koala_cube.py but need to be implemented here. #TODO
    
    Returns
    -------
    None.

    """    
    pass
# #-----------------------------------------------------------------------------
# %% ============================================================================
# #-----------------------------------------------------------------------------
def plot_wavelength(rss, wavelength, r = False, **kwargs):
    """
    Plot cut at a particular wavelength.

    Parameters
    ----------
    rss: 
        RSS object
    wavelength: int (index) or float
        wavelength to cut. If it is an integer and wavelength < n_wave, it is considered and index.
    r: Boolean    
        If True, return cut
    
    Example
    -------
    >>> plot_wavelength(rss, 7000, r =True)
    """
    w = rss.wavelength.value
    # Check if wavelength is a index or a wavelength
    n_wave = len(rss.wavelength)
    if wavelength < n_wave and type(wavelength) is int:  # wavelength is an index
        wave_index = wavelength
    else:
        wave_index= np.searchsorted(w,wavelength)
           
    wave = w[wave_index]       
    corte_wave = rss.intensity[:, wave_index].value
    
    if kwargs.get("plot") is None: plot = True
    
    if plot:
        x = range(len(rss.intensity))
        if kwargs.get("xlabel") is None:
            kwargs["xlabel"] = "Fibre"  
        if kwargs.get("ptitle") is None:
            kwargs["ptitle"] = "Intensity cut at " + str(round(wave,2)) + " $\mathrm{\AA}$ - index =" + str(wave_index)
        quick_plot(x, corte_wave, **kwargs)
    
    if r: return corte_wave
# #-----------------------------------------------------------------------------
# %% ============================================================================
# #-----------------------------------------------------------------------------
def get_spectrum(data_container, 
                 r = True, median=False,
                 # Parameters for RSS
                 fibre = None, list_fibre=None,   
                 # Parameters for CUBE
                 spaxel = None, spaxel_list = None, show_map = False,
                 **kwargs):
    """
    This task should replace tasks plot_spectrum and plot_combined_spectrum in rss and
    task spectrum_of_cube and spectrum_of_spaxel in cube.

    Parameters
    ----------
    data_container : TYPE
        DESCRIPTION.
    r : TYPE, optional
        DESCRIPTION. The default is False.
    median : TYPE, optional
        DESCRIPTION. The default is False.
    # Parameters for RSS             fibre : TYPE, optional
        DESCRIPTION. The default is None.
    list_fibre : TYPE, optional
        DESCRIPTION. The default is None.
    # Parameters for CUBE             spaxel : TYPE, optional
        DESCRIPTION. The default is None.
    spaxel_list : TYPE, optional
        DESCRIPTION. The default is None.
    show_map : TYPE, optional
        DESCRIPTION. The default is False.
    **kwargs : TYPE
        DESCRIPTION.

    Returns
    -------
    if r is True, return the spectrum.

    """
    
    verbose = kwargs.get('verbose', False)
    #warnings = kwargs.get('warnings', verbose)
    plot = kwargs.get('plot', False)
    ptitle = kwargs.get('ptitle', None)
    
    info = data_container.info
    wavelength = data_container.wavelength.value
    intensity =  data_container.intensity.value
    
    if str(type(data_container))[-5:-2] == "RSS":     # It is a RSS

        if fibre is not None:
            spectrum = intensity[fibre]
            if ptitle is None: kwargs["ptitle"] = "Spectrum of fibre {} in {}".format(fibre, info["name"])
            
        else:
            if list_fibre is None:  list_fibre = list(range(len(intensity)))
            value_list = [intensity[fibre] for fibre in list_fibre]
            if median:
                spectrum = np.nanmedian(value_list, axis=0)
            else:
                spectrum = np.nansum(value_list, axis=0)
            if ptitle is None:
                if len(list_fibre) == list_fibre[-1] - list_fibre[0] + 1:
                    kwargs["ptitle"] = "{} - Combined spectrum in range [ {} , {} ]".format(info["name"],list_fibre[0], list_fibre[-1])
                else:
                    kwargs["ptitle"] = "Combined spectrum using requested fibres"
            if verbose: 
                if median:
                    print("\n> Median spectrum of selected {} spaxels...".format(len(value_list)))
                else:    
                    print("\n> Integrating spectrum of selected {} spaxels...".format(len(value_list)))
        
    else:               # It is a cube
 
        if spaxel is None and spaxel_list is None:          # If these are not provided, plot integrated spectrum
        
            if median:
                if verbose: print("\n> Computing the median spectrum of the cube...")
                spectrum=np.nanmedian(np.nanmedian(intensity, axis=1),axis=1)
                if ptitle is None : kwargs['ptitle'] = "Median spectrum in {}".format(info["name"])
            else:
                if verbose: print("\n> Computing the integrated spectrum of the cube...")
                spectrum=np.nansum(np.nansum(intensity, axis=1),axis=1)
                if ptitle is None: kwargs['ptitle'] = "Integrated spectrum in {}".format(info["name"])
        
        elif spaxel_list is not None:
            
            n_spaxels = len(spaxel_list)
            if verbose: 
                if median:
                    print("\n> Median spectrum of selected {} spaxels...".format(n_spaxels))
                else:    
                    print("\n> Integrating spectrum of selected {} spaxels...".format(n_spaxels))
            list_of_spectra=[]        
            for i in range(n_spaxels):
                #if verbose: print("  Adding spaxel  {} : [ {}, {} ]".format(i+1,spaxel_list[i][0],spaxel_list[i][1]))
                list_of_spectra.append( [intensity[:,spaxel_list[i][1], spaxel_list[i][0]] ])
                                     
            if median:                
                spectrum=np.nanmedian(list_of_spectra, axis=0)[0]
                if ptitle is None: kwargs['ptitle'] = "Median spectrum adding {} spaxels in {}".format(n_spaxels, info["name"])
            else:    
                spectrum=np.nansum(list_of_spectra, axis=0)[0]
                if ptitle is None: kwargs['ptitle'] = "Integrated spectrum adding {} spaxels in {}".format(n_spaxels, info["name"])
        
        else:
            x,y = spaxel[1],spaxel[0]   
            spectrum = intensity[:,y,x]
            if ptitle is None : kwargs['ptitle'] = "Spaxel [{},{}] in {}".format(x,y, info["name"])
        
        if show_map: plot_map(data_container, spaxel = spaxel, spaxel_list=spaxel_list, **kwargs)
  
    if plot:        
        quick_plot(wavelength, spectrum,
                  #ptitle = ptitle,
                  #vlines=[kinfo["valid_wave_min"],kinfo["valid_wave_max"]],
                  **kwargs)    
    
    if r: return spectrum 

#-----------------------------------------------------------------------------
#-----------------------------------------------------------------------------
def plot_spectrum(rss, spectrum_number, r = False, **kwargs):
    """
    Plot spectrum of a particular fibre.

    Parameters
    ----------
    rss: 
        RSS object
    spectrum_number: int
        fibre to show spectrum.
    r: Boolean    
        If True, return spectrum
    
    Example
    -------
    >>> plot_spectrum(rss, 550)
    """
    wavelength = rss.wavelength
    spectrum = rss.intensity[spectrum_number]
    if kwargs.get("ptitle") is None:
        kwargs["ptitle"] = rss.info["name"] + " - Spectrum: "+str(spectrum_number)
    quick_plot(wavelength, spectrum, **kwargs)
    if r: return spectrum
# -----------------------------------------------------------------------------
# -----------------------------------------------------------------------------
def plot_combined_spectrum(rss, list_spectra=None,  median=False, r = False, **kwargs):
    """
    Plot combined spectrum of a list and return the combined spectrum.

    Parameters
    ----------
    rss: 
        RSS object
    list_spectra:
        spaxels to show combined spectrum. Default is all.
    median : boolean (default = False)
        if True the combined spectrum is the median spectrum
        if False the combined spectrum is the sum of the list of spectra
    r: Boolean    
        If True, return spectrum
    
    Example
    -------
    >>> star_spectrum = plot_spectrum(rss, list_spectra= [550:550], r = True)
    """
    if list_spectra is None:  list_spectra = list(range(len(rss.intensity)))

    value_list = [rss.intensity[fibre] for fibre in list_spectra]
    
    if median:
        spectrum = np.nanmedian(value_list, axis=0)
    else:
        spectrum = np.nansum(value_list, axis=0)

    plot= kwargs.get("plot", True)
 
    if plot:
        #vlines = [self.valid_wave_min, self.valid_wave_max]   #TODO  # It would be good to indicate the good wavelength range, but now these are not saved in RSS object
        if kwargs.get("ptitle") is None:
            if len(list_spectra) == list_spectra[-1] - list_spectra[0] + 1:
                kwargs["ptitle"] = "{} - Combined spectrum in range [{},{}]".format(rss.info["name"],list_spectra[0], list_spectra[-1])
            else:
                kwargs["ptitle"] = "Combined spectrum using requested fibres"
        quick_plot(rss.wavelength, spectrum,  **kwargs)
                  
    if r: return spectrum       
# -----------------------------------------------------------------------------
# -----------------------------------------------------------------------------
def rss_image(rss=None, image=None, log=False, gamma=0,
              cmap="seismic_r", clow=None, chigh=None, percentile_min = 5, percentile_max=95,
              greyscale = False, greyscale_r = False,
              xmin = None, xmax= None, wmin = None, wmax= None, fmin=0, fmax=None, 
              title_fontsize=12, title=None, add_title = False,
              xlabel = "Wavelength vector", ylabel = "Fibre", 
              axes_label_fontsize=10, axes_ticksize=None, axes_fontsize=10, axes_thickness=0, 
              colorbar_label_fontsize = 10, colorbar_ticksize= 10, colorbar_rotation = 270,  
              colorbar_width_fraction=None, colorbar_pad=None, colorbar_labelpad=10,
              colorbar_label="Intensity [Arbitrary units]", 
              fig_size=8, save_plot_in_file = None, 
              **kwargs):
    """
    Plot image of rss coloured by variable. 
    By default the cmap is "seismic_r" using symetric scale [-max_abs, max_abs],
    where max_abs = np.nanmax([np.abs(clow), np.abs(chigh)])
    
    Try greyscale=True or cmap = "binary_r" for a nice greyscale map.

    Parameters
    ----------
    rss : object
        object with the rss information.
    image : string (default = none)
        Specify the name of saved RSS image. If not, it uses rss.intensity
    log and gamma:
        Normalization scale, default is lineal scale.
        Lineal scale: norm=colors.Normalize().
        Log scale:    norm=colors.LogNorm()
        Power law:    norm=colors.PowerNorm(gamma=1./4.)  if gamma given
    log : Boolean, optional
        Plot in log scale. The default is False.
    gamma : float, optional
        Plot using Power law if gamma > 0. The default is 0.
    cmap : string 
        Colour map for the plot. The default is "seismic_r".
    clow : float, optional
        Lower bound, if not specified, 5th percentile is chosen. The default is None.
    chigh : float, optional
        Higher bound, if not specified, 95th percentile is chosen. The default is None.
    greyscale : Boolean, optional
        Use nice cmap = "binary_r" for greyscale. The default is False.
    xmin : integer, optional
        minimum value of x to plot. The default is None, that is [0]
    xmax : integer, optional
        maximum value of x to plot. The default is None, that is [-1]
    wmin : float, optional
        minimum value of wavelength to plot. If both wmin and wmax are given, the x-axis is shown in Angstroms. The default is None.
    wmax : float, optional
        maximum value of wavelength to plot. If both wmin and wmax are given, the x-axis is shown in Angstroms. The default is None.
    fmin : integer, optional
        minimum fibre to plot. The default is 0.
    fmax : integer, optional
        maximum fibre to plot. The default is None, that is the last fibre
    title_fontsize : float, optional
        Size of the font in title. The default is 12.
    title : string, optional
        Title of the plot. The default is rss.info['description'] + " - RSS image"
    xlabel : string, optional
        Label of the x-axis. The default is "Wavelength vector", but if both wmin and wmax are given, it changes to "Wavelength [$\mathrm{\AA}$]".
    ylabel : string, optional
        Label of the y-axis. The default is "Fibre".
    axes_label_fontsize : float, optional
        Size of the font of the labels in axes. The default is 10.
    axes_ticksize : list of 4 float, optional
        Size of the ticks of the axes. The default is None, that is [10,1,5,1]
    axes_fontsize : float, optional
        Size of the font of the axes. The default is 10.
    axes_thickness : float, optional
        Thickness of the axes. The default is 0.
    colorbar_label_fontsize : float, optional
        Size of the label in the colorbar. The default is 10.
    colorbar_ticksize : float, optional
        Size of the ticks in the colorbar. The default is 14.
    colorbar_rotation : float, optional
        Rotation of the colorbar. The default is 270.
    colorbar_width_fraction : float, optional
        Fraction of the plot used for the width of the colorbar. The default is None, that is 0.03 or using aspect_ratio=len(x)/len(y)
    colorbar_pad : float, optional
        Space between main figure an colorbar, in a fraction of figure. The default is None, that is 0.01, or using aspect_ratio=len(x)/len(y)
    colorbar_labelpad : float, optional
        Space between the colorbar and its label. The default is 10.
    colorbar_label : String, optional
        Label of the colorbar. The default is "Intensity [Arbitrary units]".
    fig_size : float or string, optional
        Size of the figure. If "big" it will used a specified big size. The default is 8. 
    save_plot_in_file : string, optional
        filename of the image to be saved. If not None, the image will be saved. Include ".png", ".eps" or ".pdf" at the end to be sure it is saved in the desired format. The default is None.
    **kwargs : dictionary
        list of kwargs.

    Returns
    -------
    None.
    """
    verbose = kwargs.get('verbose', False)
    #warnings = kwargs.get('warnings', verbose)
    #plot =  kwargs.get('plot', False)
    
    if rss is None and image is None: raise RuntimeError("No rss or image provided!!")
    
    # Check if image is given
    if image is None:
        image = rss.intensity
        
    # Check if rss is given
    if rss is None:
        wavelength = np.arange(len(image[0]))
        name = "IMAGE "
    else:
        wavelength = rss.wavelength
        name = rss.info['name']
        
    # Check color visualization
    norm=colors.LogNorm()
    if log is False: norm = colors.Normalize()
    if gamma > 0: norm=colors.PowerNorm(gamma=gamma)
    if greyscale: cmap = "binary_r" 
    if greyscale_r: cmap = "binary" 

    #  Set color scale and map   
    if clow is None:
        clow = np.nanpercentile(image, percentile_min)
    if chigh is None:
        chigh = np.nanpercentile(image, percentile_max)
    if cmap == "seismic_r" and log == False:
        max_abs = np.nanmax([np.abs(clow), np.abs(chigh)])
        clow = -max_abs
        chigh = max_abs
    if log and clow <=0:
        clow_=clow
        clow = np.nanmin(np.abs(image))
        if clow == 0 : 
            if chigh > 100 : 
                clow = 1.0
            else:
                clow = 0.0001
        if verbose: print("\n> Plotting image in log but the lowest value is {:.2f}, using the minimum positive value of {:.2e} instead.".format(clow_,clow))  
    
    # Show only a subregion if requested
    if xmin is None:
        xmin = 0
    if xmax is None:
        xmax = len(wavelength) 
    if wmin is not None:
        xmin = np.searchsorted(wavelength, wmin)
        if xmin == 0 : wmin = wavelength[0]
    if wmax is not None:
        xmax = np.searchsorted(wavelength, wmax)
        if xmax == len(wavelength)-1 : wmax = wavelength[-1]
    if wmin is not None and wmax is not None:
        extent1 = wmin
        extent2 = wmax
        if xlabel == "Wavelength vector": xlabel="Wavelength [$\mathrm{\AA}$]"
    else:
        extent1 = xmin
        extent2 = xmax
        
    if fmax is None: 
        fmax = len(image)
    else:
        if fmax > len(image) - 1 : fmax = len(image)
    
    extent3 = fmax
    extent4 = fmin    
    image= image[fmin:fmax,xmin:xmax]    
    
    # Check if we are saving in file to increase fig size
    if save_plot_in_file is not None:
        if type(fig_size) is not str: fig_size = fig_size * 3
        title_fontsize = title_fontsize *2.1
        axes_label_fontsize= axes_label_fontsize *2.1
        axes_fontsize=axes_fontsize *1.7
        colorbar_label_fontsize = colorbar_label_fontsize * 1.8
        colorbar_ticksize = colorbar_ticksize * 2.8
        axes_thickness = 2
        if axes_ticksize == None: axes_ticksize = [10,1,5,1]

    if type(fig_size) is str: 
        if fig_size == "big":
            fig_size=20
            title_fontsize=22
            axes_label_fontsize=18
            axes_fontsize=15
            colorbar_label_fontsize = 15
            colorbar_ticksize = 14
            axes_ticksize=[10,1,5,1]
            axes_thickness =2
     
    if axes_ticksize == None: axes_ticksize = [5,1,2,1]
    
    # Compute aspect ratio & prepare plot
    aspect_ratio = (xmax-xmin) / (fmax-fmin)
    if aspect_ratio > 1:
        fig, ax = plt.subplots(figsize=(fig_size, fig_size*aspect_ratio))
    else:        
        fig, ax = plt.subplots(figsize=(fig_size/aspect_ratio, fig_size))        
    
    # Create plot including minorticks
    pmapa=ax.imshow(image, norm=norm, cmap=cmap, clim=(clow, chigh), extent=(extent1, extent2, extent3, extent4))    
    plt.gca().invert_yaxis()    
    plt.minorticks_on()
    
    # Include labels
    plt.xlabel(xlabel, fontsize=axes_label_fontsize)
    plt.ylabel(ylabel, fontsize=axes_label_fontsize)    
    plt.tick_params('both', length=axes_ticksize[0], width=axes_ticksize[1], which='major')
    plt.tick_params('both', length=axes_ticksize[2], width=axes_ticksize[3], which='minor')
    plt.tick_params(labelsize=axes_fontsize)
    
    # Include title
    if title is not None:
        if add_title:
            plt.title(name + str(title), fontsize=title_fontsize)
        else:
            plt.title(str(title), fontsize=title_fontsize)
    else:
        plt.title(name + " - RSS image", fontsize=title_fontsize)
        
    # Making axes thicker if requested
    if axes_thickness > 0:
        plt.axhline(y=fmin,linewidth=axes_thickness, color="k")     
        plt.axvline(x=xmin,linewidth=axes_thickness, color="k")    
        plt.axhline(y=fmax,linewidth=axes_thickness, color="k")    
        plt.axvline(x=xmax,linewidth=axes_thickness, color="k")   

    # # Include colorbar well aligned with the map    
    if 0.8  <= aspect_ratio <= 1.2:
        bth = 0.05
        gap = 0.03
    else:
        bth = 0.03
        gap = 0.02    
    if colorbar_width_fraction is None: colorbar_width_fraction = np.nanmin((0.03, np.max((aspect_ratio * bth, bth))))
    if colorbar_pad is None: colorbar_pad = np.nanmin((0.01, np.max((aspect_ratio * gap, gap))))
    cax  = ax.inset_axes((1+colorbar_pad, 0, colorbar_width_fraction , 1))
    cbar = fig.colorbar(pmapa,cax=cax) 
    cbar.ax.tick_params(labelsize=colorbar_ticksize)
    cbar.set_label(str(colorbar_label), rotation=colorbar_rotation, labelpad=colorbar_labelpad, fontsize=colorbar_label_fontsize)


    # Show or save final plot
    if save_plot_in_file is None:
            plt.show()
            plt.close() 
    else:
       #if path != "" : save_file=full_path(save_file,path)
       plt.savefig(save_plot_in_file, bbox_inches='tight')
       plt.close() 
       if verbose: print("  Figure saved in file",save_plot_in_file)


# -----------------------------------------------------------------------------
# -----------------------------------------------------------------------------
def compare_rss_images(rss, image_list=None, log=False, gamma=0,
              cmap="seismic_r", clow=None, chigh=None, percentile_min = 5, percentile_max=95,
              greyscale = False, greyscale_r = False,
              xmin = None, xmax= None, wmin = None, wmax= None, fmin=0, fmax=None, 
              title_fontsize=12, title=None, add_title = False,
              xlabel = "Wavelength vector", ylabel = "Fibre", 
              axes_label_fontsize=10, axes_ticksize=None, axes_fontsize=10, axes_thickness=0, 
              colorbar_label_fontsize = 10, colorbar_ticksize= 10, colorbar_rotation = 270,  
              colorbar_width_fraction=None, colorbar_pad=None, colorbar_labelpad=10,
              colorbar_label="Intensity [Arbitrary units]", 
              fig_size=[8,6], save_plot_in_file = None, 
              **kwargs):
    verbose = kwargs.get('verbose', False)
    #warnings = kwargs.get('warnings', verbose)
    #plot =  kwargs.get('plot', False)
    
    # Show only a subregion if requested
    if xmin is None:
        xmin = 0
    if xmax is None:
        xmax = len(rss.wavelength) 
    if wmin is not None:
        xmin = np.searchsorted(rss.wavelength, wmin)
        if xmin == 0 : wmin = rss.wavelength[0]
    if wmax is not None:
        xmax = np.searchsorted(rss.wavelength, wmax)
        if xmax == len(rss.wavelength)-1 : wmax = rss.wavelength[-1]
    if wmin is not None and wmax is not None:
        extent1 = wmin
        extent2 = wmax
        if xlabel == "Wavelength vector": xlabel="Wavelength [$\mathrm{\AA}$]"
    else:
        extent1 = xmin
        extent2 = xmax
    if fmax is None: 
        fmax = len(rss.intensity)
    else:
        if fmax > len(rss.intensity) - 1 : fmax = len(rss.intensity)
    extent3 = fmax
    extent4 = fmin    
    
    # Check if we are saving in file to increase fig size
    if save_plot_in_file is not None:
        if type(fig_size) is not str: fig_size = fig_size * 3
        title_fontsize = title_fontsize *2.1
        axes_label_fontsize= axes_label_fontsize *2.1
        axes_fontsize=axes_fontsize *1.7
        colorbar_label_fontsize = colorbar_label_fontsize * 1.8
        colorbar_ticksize = colorbar_ticksize * 2.8
        #axes_thickness = 2
        if axes_ticksize == None: axes_ticksize = [10,1,5,1]

    if type(fig_size) is str: 
        if fig_size == "big":
            fig_size=20
            title_fontsize=22
            axes_label_fontsize=18
            axes_fontsize=15
            colorbar_label_fontsize = 15
            colorbar_ticksize = 14
            axes_ticksize=[10,1,5,1]
            #axes_thickness =2
     
    if axes_ticksize == None: axes_ticksize = [5,1,2,1]


    # Check position of colorbars
    if colorbar_pad is None: colorbar_pad =0.05
    if colorbar_width_fraction is None: colorbar_width_fraction =0.15
    
    # Start figure with sublots   
    a = 1   # All horizontal
    n_plots = len (image_list)
    fig, axs = plt.subplots(a, n_plots , figsize=(fig_size[0], fig_size[1]),
                            constrained_layout=True, sharey=True,
                            sharex=True)
    
    if type(greyscale) is bool: greyscale = [greyscale]*n_plots
    if type(greyscale_r) is bool: greyscale_r = [greyscale_r]*n_plots  
    if type(gamma) is int or float : gamma = [gamma] *n_plots 
    
    i = 0  # Looping the plots
    for col in range(n_plots):
        #for row in range(a):  #### IF in the future we need to expand this...
            
        data = image_list[i][fmin:fmax,xmin:xmax]            
        
        # Check color visualization
        norm=colors.LogNorm()
        if log[i] is False: norm = colors.Normalize()
        if gamma[i] > 0: norm=colors.PowerNorm(gamma=gamma[i])
        if greyscale[i]: cmap[i] = "binary_r" 
        if greyscale_r[i]: cmap[i] = "binary" 
    
        #  Set color scale and map   
        if clow[i] is None:
            clow[i] = np.nanpercentile(data, percentile_min)
        if chigh[i] is None:
            chigh[i] = np.nanpercentile(data, percentile_max)
        if cmap[i] == "seismic_r" and log[i] == False:
            max_abs = np.nanmax([np.abs(clow[i]), np.abs(chigh[i])])
            clow[i] = -max_abs
            chigh[i] = max_abs
        if log[i] and clow[i] <=0:
            clow[i] = np.nanmin(np.abs(data))
            if clow[i] == 0 : 
                if chigh[i] > 100 : 
                    clow[i] = 1.0
                else:
                    clow[i] = 0.0001
          
        # Create subplot including minorticks
        ax = axs[col]
        pcm=ax.imshow(data, norm=norm, cmap=cmap[i], clim=(clow[i], chigh[i]), extent=(extent1, extent2, extent3, extent4))    
        plt.minorticks_on()
        plt.gca().invert_yaxis()   # Invert axis in y 
        
        plt.tick_params('both', length=axes_ticksize[0], width=axes_ticksize[1], which='major')
        plt.tick_params('both', length=axes_ticksize[2], width=axes_ticksize[3], which='minor')
        
        # Include labels
        ax.set_xlabel(xlabel,  fontsize=13)
        ax.set_ylabel(ylabel, fontsize=13) 
        ax.set_title(title[i])  ##, fontsize=title_fontsize)
        
        # Color bar
        cax  = ax.inset_axes((1+colorbar_pad, 0, colorbar_width_fraction , 1))
        cbar = fig.colorbar(pcm,cax=cax) 
        #cbar.ax.tick_params(labelsize=colorbar_ticksize)
        if colorbar_label[i] is not None: cbar.set_label(str(colorbar_label[i])) #, rotation=colorbar_rotation, labelpad=colorbar_labelpad, fontsize=colorbar_label_fontsize)

        i=i+1 # Prepare for next plot

    
            
    # Show or save final plot
    if save_plot_in_file is None:
            plt.show()
            plt.close() 
    else:
       #if path != "" : save_file=full_path(save_file,path)
       plt.savefig(save_plot_in_file, bbox_inches='tight')
       plt.close() 
       if verbose: print("  Figure saved in file",save_plot_in_file)


# -----------------------------------------------------------------------------
# -----------------------------------------------------------------------------
def rss_map(rss, variable=None, list_spectra=None, 
            instrument = None,
            clow=None, chigh=None,
            log=True, floor = None, gamma=0, cmap = None,
            title=" - RSS map",  description = None,
            xlabel = None, ylabel = None, 
            color_bar_text="Integrated Flux [Arbitrary units]", 
            figsize = 8, grid=False,
            colorbar_labelpad = 10, colorbar_pad= None, colorbar_width_fraction= None,
            save_plot_in_file = None,
            **kwargs):
    """
    Plot map showing the offsets, coloured by variable.

    Parameters
    ----------
    norm:
        Normalization scale, default is lineal scale.
        Lineal scale: norm=colors.Normalize()
        Log scale:    norm=colors.LogNorm()
        Power law:    norm=colors.PowerNorm(gamma=1./4.)
    list_spectra : list of floats (default = none)
        List of RSS spectra
    title : string (default = " - RSS image")
        Set plot title
    color_bar_text : string (default = "Integrated Flux [Arbitrary units]")
        Colorbar's label text
    """
    verbose = kwargs.get('verbose', False)
    warnings = kwargs.get('warnings', verbose)

    if variable is None: 
        variable, _ = rss.get_integrated_fibres()

    if cmap is None:
        cmap = fuego_color_map

    if gamma > 0: 
        norm=colors.PowerNorm(gamma=gamma)
    elif log:
        norm=colors.LogNorm()     
        min_value = np.nanmin(variable)
        if min_value <= 0:
            if floor is None:
                floor = np.nanmin(np.abs(variable))
                if floor == 0: floor = 0.1
                variable = np.array([np.nan if x <= floor else x for x in variable])
                negative_spaxels = np.count_nonzero(np.isnan(variable))
                if warnings: print(f"> WARNING: The image has {negative_spaxels} spaxels with <= 0 values, min_value = {min_value}, using nan for these...")
            else:
                variable = np.array([floor if x <= floor else x for x in variable])  
                negative_spaxels = variable.tolist().count(floor)
                if warnings: print(f"> WARNING: The image has {negative_spaxels} spaxels with <= 0 values, min_value = {min_value}, using provided floor = {floor} for these...")
            if clow is None: clow = floor
    else:
        norm = colors.Normalize()
        
    if clow is None: clow = np.nanmin(variable[list_spectra])
    if chigh is None: chigh = np.nanmax(variable[list_spectra])

    if list_spectra is None: list_spectra = list(range(len(rss.intensity)))

    # Get coordinates
    fib_ra = rss.info["fib_ra"]
    fib_dec = rss.info["fib_dec"]
    
    RA_cen = fib_ra.mean()
    DEC_cen = fib_dec.mean()
    
    offset_RA_arcsec = (fib_ra - RA_cen) * 3600.
    offset_DEC_arcsec = (fib_dec - DEC_cen) * 3600.

    # Start figure
    if np.isscalar(figsize) :
        fig, ax = plt.subplots(figsize=(figsize, figsize))
        aspect_ratio = 1.
    else:
        fig, ax = plt.subplots(figsize=(figsize[0], figsize[1]))
        aspect_ratio = figsize[0] / figsize[1]
        figsize = figsize[0]
    
    cax= ax.scatter(offset_RA_arcsec[list_spectra],
                offset_DEC_arcsec[list_spectra],
                c=variable[list_spectra], cmap=cmap, 
                clim=(clow, chigh),
                norm=norm,
                s=20*figsize, marker="h")
    if description is None:
        ax.set_title(rss.info["name"] + title)   
    else:
        ax.set_title(str(description))   # description
    ax.set_facecolor("lightgrey")
    
    plt.xlim(np.nanmin(offset_RA_arcsec) - 0.75, np.nanmax(offset_RA_arcsec) + 0.75)
    plt.ylim(np.nanmin(offset_DEC_arcsec) - 0.75, np.nanmax(offset_DEC_arcsec) + 0.75)
    plt.gca().invert_xaxis()   # RA is NEG to the right
    if xlabel is None: xlabel ="$\Delta$ RA [arcsec]" 
    if ylabel is None: ylabel="$\Delta$ DEC [arcsec]"
    plt.xlabel(str(xlabel))
    plt.ylabel(str(ylabel))    
    plt.minorticks_on()
    if grid: plt.grid(which='both')
    
    # Set colorbar
    if colorbar_pad is None : colorbar_pad = 0.02
    if colorbar_width_fraction is None: colorbar_width_fraction = 0.4        
    cbar = fig.colorbar(cax, pad = colorbar_pad, aspect = figsize*aspect_ratio / colorbar_width_fraction ) #, width=figsize*colorbar_width_fraction) # colorbar_width_fraction
    cbar.set_label(str(color_bar_text), rotation=90, labelpad=colorbar_labelpad)
    cbar.ax.tick_params()
    
    # Show or save final plot
    if save_plot_in_file is None:
            plt.show()
            plt.close() 
    else:
       plt.savefig(save_plot_in_file, bbox_inches='tight')
       plt.close() 
       if verbose: print("  Figure saved in file",save_plot_in_file, **kwargs)



# Ángel :-)


from ._version import get_versions
import logging
import sys

__version__ = get_versions()['version']
del get_versions

# Parent logger
pykoala_logger = logging.getLogger('pykoala')

if not (pykoala_logger.hasHandlers()):
    stdout = logging.StreamHandler(sys.stdout)
    fmt = logging.Formatter(
    "[%(name)s] %(asctime)s|%(levelname)s> %(message)s",
    datefmt="%Y/%m/%d %H:%M")
    stdout.setFormatter(fmt)
    pykoala_logger.addHandler(stdout)
    pykoala_logger.setLevel(logging.INFO)

def log_into_file(filename, logger_name='pykoala', level='INFO'):
    logger = logging.getLogger(logger_name)
    hdlr = logging.FileHandler(filename)
    fmt = logging.Formatter(
    "[%(name)s] %(asctime)s|%(levelname)s> %(message)s",
    datefmt="%Y/%m/%d %H:%M")
    hdlr.setFormatter(fmt)
    logger.addHandler(hdlr) 
    logger.setLevel(level.upper())


def vprint(msg, *args, **kwargs):
    """
    Convenience function for using with the pykoala generic logger.
    """
    logger = logging.getLogger('pykoala')
    print_method = getattr(logger, kwargs.get('level', 'info').lower())
    print_method(msg, *args, **kwargs)


class VerboseMixin():

    @property
    def logger(self):
        return self._logger

    @logger.setter
    def logger(self, logger):
        if isinstance(logger, str):
            if logger == pykoala_logger.name:
                self._logger = pykoala_logger
            else:
                self._logger = pykoala_logger.getChild(logger)
        elif isinstance(logger, logging.Logger):
            self._logger = logger

    @property
    def verbose(self):
        """Abstract property for the verbosity of the correction."""
        return self._verbose

    @verbose.setter
    def verbose(self, verbose):
        if verbose:
            self.logger.setLevel(logging.INFO)
        else:
            self.logger.setLevel(logging.ERROR)
        self._verbose = verbose
    
    def vprint(self, msg, level='info'):
        """
        Logs a message at the specified level.

        Parameters
        ----------
        msg : str
            The message to be logged.
        level : str, optional
            The level at which to log the message (default is 'info').
        """
        printer = getattr(self.logger, level)
        printer(msg)

    def log_into_file(self, filename, level="INFO"):
        log_into_file(filename, self.logger.name, level=level)


__all__ = ["__version__", "pykoala_logger", "log_into_file", "vprint",
           "VerboseMixin"]
from . import _version
__version__ = _version.get_versions()['version']
